#!/private/tmp/venv/bin/python
# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2016.                            (c) 2016.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import netrc
import argparse
import os
import signal
import sys
from six.moves.urllib.parse import urlparse

from .net.auth import get_cert, CRED_RESOURCE_ID, Subject
from .net.ws import BaseWsClient, SERVICE_AVAILABILITY_ID
from cadcutils import exceptions

# CADC realms current and old
CADC_REALMS = ['www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca',
               'www.canfar.phys.uvic.ca']
"""

The code in this file is deprecated being replaced by certificate related code
in the net.auth module.
It is kept here only to support users that still rely on getCert application

"""


def _main():

    def _signal_handler(signal, frame):
        sys.stderr.write("\n")
        sys.exit(-1)

    signal.signal(signal.SIGINT, _signal_handler)

    # Note: we make the assumption that the CDP end point and its availability
    # end point are on the same host. We build a dummy anonymous client to
    # the service and retrieve the access url of the availability end point
    # which then is used to determine the actual realm of the service
    dummy = BaseWsClient(CRED_RESOURCE_ID, Subject(),
                         agent="getCert/1.0", retry=True)
    access_url = dummy._get_url((SERVICE_AVAILABILITY_ID, None))
    service_realm = urlparse(access_url).netloc

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=("Retrieve a security certificate for interaction with "
                     "VOSpace. Certificate will be valid for daysValid and "
                     "stored as local file cert_filename. First looks for an "
                     "entry in the users .netrc matching the realm {0}, "
                     "the user is prompted for a username and "
                     "password if no entry is found.".format(service_realm)))

    parser.add_argument('--version', '-V', action='version', version='1.1')
    parser.add_argument('--daysValid', type=int, default=10,
                        help='Number of days the cetificate should be valid.')
    parser.add_argument('--cert-filename', default=os.path.join(
            os.getenv('HOME'), '.ssl', 'cadcproxy.pem'),
                        help="Filesysm location to store the proxy "
                             "certifcate.")
    parser.add_argument('--cert-server', default=None,
                        help="Deprecated. Used for testing only")

    args = parser.parse_args()
    if args.cert_server:
        # update service_realm
        dummy = BaseWsClient(CRED_RESOURCE_ID, Subject(),
                             agent="getCert/1.0", retry=True,
                             host=args.cert_server)
        access_url = dummy._get_url((SERVICE_AVAILABILITY_ID, None))
        service_realm = urlparse(access_url).netloc
        print('Service realm: {}'.format(service_realm))

    if os.access(os.path.join(os.environ.get('HOME', '/'), ".netrc"), os.R_OK):
        # try the service_realm first and if that is not found in the
        # .netrc file, try any of the CADC equivalents
        auth = netrc.netrc().authenticators(service_realm)
        if not auth:
            for r in CADC_REALMS:
                auth = netrc.netrc().authenticators(r)
                if auth:
                    break
    else:
        auth = False
    if not auth:
        sys.stdout.write("{0} Username: ".format(service_realm))
        sys.stdout.flush()
        username = sys.stdin.readline().strip('\n')
        subject = Subject(username=username)
    else:
        subject = Subject(netrc=True)
        # the following prevents reading the .netrc file and uses the
        # auth already discovered.
        subject._hosts_auth[service_realm] = (auth[0], auth[2])

    retry = True
    while retry:
        try:
            cert = get_cert(subject, days_valid=args.daysValid)
            retry = False
            with open(args.cert_filename, 'w') as w:
                w.write(cert)
        except exceptions.UnauthorizedException:
            sys.stderr.write("FAILED: invalid username/password combination")
            return -1
        except OSError as ose:
            if ose.errno != 401:
                sys.stderr.write(str(ose))
                return getattr(ose, 'errno', 1)
            else:
                sys.stderr.write("Access denied\n")
        except Exception as ex:
            sys.stderr.write(str(ex))
            return getattr(ex, 'errno', 1)


if __name__ == '__main__':
    sys.exit(_main())



# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2021.                            (c) 2021.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

"""
Module that contains functionality related to CADC groups.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import re

from cadcutils import version  # TODO should it be application version instead?

from cadcutils import util

CADC_LOGIN_CAPABILITY = 'ivo://ivoa.net/std/UMS#login-0.1'
CADC_SSO_COOKIE_NAME = 'CADC_SSO'
CADC_REALMS = ['.canfar.net', '.cadc-ccda.hia-iha.nrc-cnrc.gc.ca',
               '.cadc.dao.nrc.ca']

# ID of the default catalog Web service
CADC_AC_SERVICE_ID = 'ivo://cadc.nrc.ca/gms'
DEFAULT_SERVICE_ID = 'ivo://cadc.nrc.ca/gms'

SEARCH_CAPABILITY_ID = 'ivo://ivoa.net/std/gms#search-1.0'

# spec extensions relative to root of GMS service URL
GROUPS_PATH = 'groups'
GROUP_MEMBER_PATH = 'groupMembers'
USER_MEMBER_PATH = 'userMembers'


__all__ = ['User', 'Group', 'Role', 'GroupProperty', 'Identity']

# these are the security methods currently supported
SECURITY_METHODS_IDS = {
    'certificate': 'ivo://ivoa.net/sso#tls-with-certificate',
    'cookie': 'ivo://ivoa.net/sso#cookie'}

logger = util.get_logger(__name__)

cadcgms_agent = 'cadc-gms-client/{}'.format(version.version)


class User:
    """
    Class representing a CADC groups user
    """

    def __init__(self, internal_id=None):
        """
        internal_id is a uri which uniquely identifies the user.
        The uri scheme and scheme-specific-part uniquely identify the service,
        while the uri fragment uniquely identifies the user of that service.
        """
        self.internal_id = internal_id
        self.identities = {}

    def __eq__(self, other):
        return self.internal_id == other.internal_id

    def __hash__(self):
        return hash(self.internal_id)


class Role:
    """Helper class for constraining allowable roles"""

    allowable_roles = set(['owner', 'member', 'admin'])

    def __init__(self, role_name):
        name_lower = role_name.lower()
        if name_lower in self.allowable_roles:
            self._role = name_lower
        else:
            raise ValueError("Role %s is not in the allowable list: %s" % (
                name_lower, self.allowable_roles))

    def get_name(self):
        return self._role

    def __eq__(self, other):
        return self._role == other._role

    def __hash__(self):
        return hash(self._role)


class Identity():

    identity_types = ['X500', 'OpenID', 'HTTP', 'CADC', 'POSIX']

    def __init__(self, name, identity_type):
        if name is None or not name:
            raise ValueError('Identity name is None or empty')
        if identity_type not in self.identity_types:
            raise ValueError('Unknown Identity type ' + identity_type)
        self.name = name
        self.type = identity_type

    def __eq__(self, other):
        return self.name, self.type == other.name, other.type


class GroupProperty:

    STRING_TYPE = 'String'
    INTEGER_TYPE = 'Integer'

    def __init__(self, key, value, read_only=False):
        """
        A property representing metadata for a group.
        """
        if key is None:
            raise ValueError('GroupProperty key cannot be None')
        if value is None:
            raise ValueError('GroupProperty value cannot be None')

        self.key = key
        self.value = value
        self.read_only = read_only

    def __eq__(self, other):
        return self.key, self.value, self.read_only == other.key, other.value,\
               other.read_only

    def __hash__(self):
        return hash((self.key, self.value))


class Group(object):
    """
    Class representing a CADC group
    """

    def __init__(self, group_id, authority='ivo://cadc.nrc.ca/gms'):
        """
        Ctor
        :param group_id: Group name.
        :param authority: The authority under which this group name is
        registered
        """
        group_id_re = re.compile('^[a-zA-Z0-9\\-\\.~_]*$')
        if not group_id_re.match(group_id):
            raise Exception(
                'Invalid group ID {} may not contain space ( ), slash (/), '
                'escape (\\), or percent (%)'.format(group_id_re))

        self.group_id = group_id
        self.authority = authority
        self.uri = '{}?{}'.format(authority, group_id)

        self.owner = None
        self.description = None
        self.last_modified = None
        self.properties = set()
        self.user_members = set()
        self.group_members = set()
        self.user_admins = set()
        self.group_admins = set()

    def add_user_member(self, user):
        """

        :param user: member to add (User type)
        """
        if not isinstance(user, User):
            raise AttributeError('user must be of `User` type')
        self.user_members.add(user)

    def add_user_admin(self, user):
        """
        Adds an admin for the group
        :param user: admin to add (User type)
        """
        if not isinstance(user, User):
            raise AttributeError('user must be of `User` type')
        self.user_admins.add(user)

    def __eq__(self, other):
        return self.group_id == other.group_id

    def __hash__(self):
        return hash(self.group_id)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2022.                            (c) 2022.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import os
import time
import unittest

import requests
from mock import Mock, patch, call, mock_open, ANY
from six import StringIO
from six.moves.urllib.parse import urlparse
import tempfile
import pytest
import hashlib
from tempfile import NamedTemporaryFile, TemporaryDirectory

from cadcutils import exceptions
from cadcutils import net
from cadcutils.net import ws, auth
from cadcutils.net.ws import DEFAULT_RETRY_DELAY, MAX_RETRY_DELAY, \
    MAX_NUM_RETRIES, SERVICE_RETRY

# The following is a temporary workaround for Python issue
# 25532 (https://bugs.python.org/issue25532)
call.__wrapped__ = None

# Content type header
CONTENT_TYPE = 'Content-Type'
TEXT_TYPE = 'text/plain'


class TestListResources(unittest.TestCase):
    @patch('cadcutils.net.ws.requests.get')
    def test_list_resources(self, get_mock):
        response_caps = Mock()
        response_caps.text = (
            '# This is just a test\n'
            'ivo://cadc.nrc.ca/serv1 = '
            'http://www.cadc.nrc.gc.ca/serv1/capabilities\n'
            'ivo://cadc.nrc.ca/serv2 = '
            'http://www.cadc.nrc.gc.ca/serv2/capabilities\n')
        response_serv1 = Mock()
        response_serv1.text = (
            '<vosi:capabilities xmlns:vosi='
            '"http://www.ivoa.net/xml/VOSICapabilities/v1.0" '
            'xmlns:vs="http://www.ivoa.net/xml/VODataService/v1.1" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\n'
            '<capability standardID="ivo://ivoa.net/std/VOSI#capabilities">\n'
            '<interface xsi:type="vs:ParamHTTP" role="std">\n'
            '<accessURL use="full">'
            'http://www.cadc.hia.nrcgc..ca/serv1/capabilities</accessURL>\n'
            '</interface>\n'
            '</capability>\n'
            '<capability standardID="ivo://ivoa.net/std/VOSI#availability">\n'
            '<interface xsi:type="vs:ParamHTTP" role="std">\n'
            '<accessURL use="full">'
            'http://www.cadc.nrc.gc.ca/serv1/availability</accessURL>\n'
            '</interface>\n'
            '</capability>\n'
            '</vosi:capabilities>\n')
        response_serv2 = Mock()
        response_serv2.text = (
            '<vosi:capabilities xmlns:vosi='
            '"http://www.ivoa.net/xml/VOSICapabilities/v1.0" '
            'xmlns:vs="http://www.ivoa.net/xml/VODataService/v1.1" '
            'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\n'
            '<capability standardID="ivo://ivoa.net/std/VOSI#capabilities">\n'
            '<interface xsi:type="vs:ParamHTTP" role="std">\n'
            '<accessURL use="full">'
            'http://www.cadc.hia.nrcgc..ca/serv2/capabilities</accessURL>\n'
            '</interface>\n'
            '</capability>\n'
            '<capability standardID="ivo://ivoa.net/std/VOSI#tables-1.1">\n'
            '<interface xsi:type="vs:ParamHTTP" role="std">\n'
            '<accessURL use="full">'
            'http://www.cadc.nrc.gc.ca/serv2/availability</accessURL>\n'
            '</interface>\n'
            '</capability>\n'
            '</vosi:capabilities>\n')
        get_mock.side_effect = [response_caps, response_serv1, response_serv2]
        self.maxDiff = None
        usage = \
            ('ivo://cadc.nrc.ca/serv1 '
             '(http://www.cadc.nrc.gc.ca/serv1/capabilities) - '
             'Capabilities: ivo://ivoa.net/std/VOSI#availability, '
             'ivo://ivoa.net/std/VOSI#capabilities\n\n'
             'ivo://cadc.nrc.ca/serv2 '
             '(http://www.cadc.nrc.gc.ca/serv2/capabilities) - '
             'Capabilities: ivo://ivoa.net/std/VOSI#capabilities, '
             'ivo://ivoa.net/std/VOSI#tables-1.1')
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            ws.list_resources()
            self.assertEqual(usage, stdout_mock.getvalue().strip())


class TestWs(unittest.TestCase):
    """Class for testing the webservie client"""

    @patch('cadcutils.net.ws.WsCapabilities')
    @patch('cadcutils.net.auth.os.path.isfile', Mock())
    @patch('cadcutils.net.auth.netrclib')
    @patch('cadcutils.net.ws.RetrySession.put')
    @patch('cadcutils.net.ws.RetrySession.head')
    @patch('cadcutils.net.ws.RetrySession.delete')
    @patch('cadcutils.net.ws.RetrySession.get')
    @patch('cadcutils.net.ws.RetrySession.post')
    def test_ops(self, post_mock, get_mock, delete_mock, head_mock, put_mock,
                 netrclib_mock, caps_mock):
        anon_subject = auth.Subject()
        with self.assertRaises(ValueError):
            ws.BaseWsClient(None, anon_subject, "TestApp")
        resource = 'aresource'
        service = 'myservice'
        resource_id = 'ivo://www.canfar.net/{}'.format(service)
        # test anonymous access
        cm = Mock()
        cm.get_access_url.return_value = "http://host/availability"
        caps_mock.return_value = cm
        client = ws.BaseWsClient(resource_id, anon_subject, 'TestApp')
        resource_uri = urlparse(resource_id)
        base_url = 'http://{}{}/pub'.format(resource_uri.netloc,
                                            resource_uri.path)
        resource_url = 'http://{}{}/{}'.format(resource_uri.netloc, base_url,
                                               resource)
        self.assertEqual(anon_subject, client.subject)
        self.assertTrue(client.retry)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        self.assertEqual(None, client._session)  # lazy initialization
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None, verify=True)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, verify=True, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url, verify=True)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url, verify=True)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, verify=True, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test basic authentication access
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        host = 'www.different.org'
        subject = auth.Subject(netrc='somecert')
        client = ws.BaseWsClient(resource_id, subject, 'TestApp', retry=False,
                                 host=host)
        base_url = 'http://{}{}/auth'.format(host, resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertFalse(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None, verify=True)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, verify=True, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url, verify=True)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url, verify=True)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, verify=True, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test cert authentication
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        certfile = 'some/certfile.pem'
        subject = auth.Subject(certificate=certfile)
        client = ws.BaseWsClient(resource_id, subject, 'TestApp')
        base_url = 'https://{}{}/pub'.format(resource_uri.netloc,
                                             resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None, verify=True)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, verify=True, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url, verify=True)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url, verify=True)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, verify=True, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))
        self.assertEqual((certfile, certfile), client._session.cert)

        # repeat above tests test with the temporary sc2repo test
        resource = 'aresource'
        service = 'sc2repo'
        resource_id = 'ivo://www.canfar.phys.uvic.ca/{}'.format(service)
        # test anonymous access
        client = ws.BaseWsClient(resource_id, auth.Subject(), 'TestApp')
        resource_uri = urlparse(resource_id)
        base_url = 'http://{}{}/observations'.format(resource_uri.netloc,
                                                     resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertTrue(client.retry)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        self.assertEqual(None, client._session)  # lazy initialization
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None, verify=True)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, verify=True, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url, verify=True)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url, verify=True)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, verify=True, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test basic authentication access
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        host = 'caom2workshop.canfar.net'
        subject = auth.Subject(netrc=True)
        subject._hosts_auth[host] = ('auser', 'apasswd')
        client = ws.BaseWsClient(resource_id, subject, 'TestApp', retry=False,
                                 host=host)
        base_url = 'http://{}{}/auth-observations'.format(host,
                                                          resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertFalse(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None, verify=True)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, verify=True, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url, verify=True)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url, verify=True)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, verify=True, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test cert authentication
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        certfile = 'some/certfile.pem'
        client = ws.BaseWsClient(resource_id,
                                 auth.Subject(certificate=certfile), 'TestApp')
        base_url = 'https://{}{}/observations'.format(resource_uri.netloc,
                                                      resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None, verify=True)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, verify=True, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url, verify=True)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url, verify=True)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, verify=True, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))
        self.assertEqual((certfile, certfile), client._session.cert)

        # test cookie authentication
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        subject = auth.Subject()
        subject.cookies.append(auth.CookieInfo(resource_uri.netloc,
                                               'MyTestCookie', 'cookievalue'))
        client = ws.BaseWsClient(resource_id, subject, 'TestApp',
                                 insecure=True)
        base_url = 'https://{}{}/observations'.format(resource_uri.netloc,
                                                      resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None, verify=False)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, verify=False, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url, verify=False)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url, verify=False)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, verify=False, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))
        self.assertEqual(1, len(client._session.cookies))
        self.assertEqual('cookievalue',
                         client._session.cookies['MyTestCookie'])

    @patch('cadcutils.net.ws.util.Md5File')
    @patch('cadcutils.net.ws.WsCapabilities')
    def test_upload_file_no_put_txn(self, caps_mock, md5_file_mock):
        anon_subject = auth.Subject()
        target_url = 'https://someurl'
        cm = Mock()
        cm.get_access_url.return_value = "http://host/availability"
        caps_mock.return_value = cm
        response = Mock()
        response.status_code = requests.codes.ok
        response.headers = {}
        session = Mock()
        session.put.return_value = response
        md5_file_mock_obj = Mock()
        md5_file_mock.return_value.__enter__.return_value = md5_file_mock_obj
        client = ws.BaseDataClient(resource_id='ivo://cadc.nrc.ca/resourceid',
                                   subject=anon_subject, agent='TestApp')
        # upload small file (md5 checksum is pre-computed)
        client._get_session = Mock(return_value=session)
        content = 'this is some content for the test'
        md5 = hashlib.md5()
        md5.update(str.encode(content))
        content_md5 = md5.hexdigest()
        src = tempfile.NamedTemporaryFile()
        with open(src.name, 'w') as f:
            f.write(content)
        md5_file_mock_obj.md5_checksum = content_md5
        net.add_md5_header(headers=response.headers, md5_checksum=content_md5)
        md5_file_mock_obj.md5_checksum = content_md5
        # add caller headers and test they are passed through
        caller_header = {CONTENT_TYPE: TEXT_TYPE}
        client.upload_file(url=target_url, src=src.name, headers=caller_header)
        session.put.assert_called_once()
        put_headers = {}
        net.add_md5_header(headers=put_headers, md5_checksum=content_md5)
        put_headers[ws.HTTP_LENGTH] = str(len(content))
        net.add_md5_header(headers=put_headers, md5_checksum=content_md5)
        expected_headers = caller_header
        expected_headers.update(put_headers)
        session.put.assert_called_with(target_url, headers=expected_headers,
                                       data=ANY, verify=True)

        # pass the md5 in update small file
        session.put.reset_mock()
        client.upload_file(url=target_url, src=src.name,
                           md5_checksum=content_md5)
        session.put.assert_called_once()
        session.put.assert_called_with(target_url, headers=put_headers,
                                       data=ANY, verify=True)

        # make it fail the first attempt but succeed on the next
        session.put.reset_mock()
        session.put.side_effect = [exceptions.PreconditionFailedException,
                                   Mock()]
        client.upload_file(url=target_url, src=src.name,
                           md5_checksum=content_md5)
        assert session.put.mock_calls == [
            call(target_url, headers=put_headers, data=ANY, verify=True),
            call(target_url, headers=put_headers, data=ANY, verify=True)]

        # fail on repeated BadRequests
        session.put.reset_mock()
        session.put.side_effect = [exceptions.PreconditionFailedException] * 4
        with pytest.raises(exceptions.PreconditionFailedException):
            client.upload_file(url=target_url, src=src.name,
                               md5_checksum=content_md5)

        # fail on other errors
        session.put.reset_mock()
        session.put.side_effect = [exceptions.UnexpectedException]
        with pytest.raises(exceptions.UnexpectedException):
            client.upload_file(url=target_url, src=src.name,
                               md5_checksum=content_md5)

    @patch('cadcutils.net.ws.util.Md5File')
    @patch('cadcutils.net.ws.WsCapabilities')
    def test_upload_file_put_txn(self, caps_mock, md5_file_mock):
        anon_subject = auth.Subject()
        target_url = 'https://someurl'
        cm = Mock()
        cm.get_access_url.return_value = "http://host/availability"
        caps_mock.return_value = cm
        response = Mock()
        response.status_code = requests.codes.ok
        response.headers = {}
        session = Mock()
        session.put.return_value = response
        md5_file_mock_obj = Mock()
        md5_file_mock.return_value.__enter__.return_value = md5_file_mock_obj
        client = ws.BaseDataClient(resource_id='ivo://cadc.nrc.ca/resourceid',
                                   subject=anon_subject, agent='TestApp')
        client._get_session = Mock(return_value=session)
        content = 'this is some content for the test'
        md5 = hashlib.md5()
        md5.update(str.encode(content))
        content_md5 = md5.hexdigest()
        src = tempfile.NamedTemporaryFile()
        with open(src.name, 'w') as f:
            f.write(content)
        md5_file_mock_obj.md5_checksum = content_md5
        net.add_md5_header(headers=response.headers, md5_checksum=content_md5)
        md5_file_mock_obj.md5_checksum = content_md5
        orig_max_md5_compute_size = ws.MAX_MD5_COMPUTE_SIZE
        try:
            # put larger one segment files
            session.put.reset_mock()
            response_headers = {ws.PUT_TXN_ID: '123',
                                ws.HTTP_LENGTH: str(len(content))}
            net.add_md5_header(response_headers, content_md5)
            session.put.return_value = Mock(headers=response_headers)
            # add caller headers and test they are passed through
            caller_header = {CONTENT_TYPE: TEXT_TYPE}
            # lower the threshold for "large" files so that the current test
            # files becomes large
            ws.MAX_MD5_COMPUTE_SIZE = 10
            # PUT headers do not contain the md5 anymore
            client.upload_file(url=target_url, src=src.name,
                               headers=caller_header)
            put_headers = {ws.HTTP_LENGTH: str(len(content)),
                           ws.PUT_TXN_OP: ws.PUT_TXN_START}
            commit_headers = {ws.PUT_TXN_ID: '123',
                              ws.PUT_TXN_OP: ws.PUT_TXN_COMMIT,
                              ws.HTTP_LENGTH: '0'}
            expected_put_headers = dict(caller_header)
            expected_put_headers.update(put_headers)
            expected_commit_headers = dict(caller_header)
            expected_commit_headers.update(commit_headers)
            assert session.put.mock_calls == \
                [call(target_url, data=ANY, verify=True,
                      headers=expected_put_headers),
                 call(target_url, verify=True,
                      headers=expected_commit_headers)]

            # repeat but provide the checksum as argument to upload_file so
            # no transaction is required
            del put_headers[ws.PUT_TXN_OP]
            session.put.reset_mock()
            client.upload_file(url=target_url, src=src.name,
                               md5_checksum=content_md5)
            net.add_md5_header(headers=put_headers, md5_checksum=content_md5)
            session.put.assert_called_with(target_url, headers=put_headers,
                                           data=ANY, verify=True)

            # mimic md5 mismatch
            session.put.reset_mock()
            response_headers = {ws.PUT_TXN_ID: '123',
                                ws.HTTP_LENGTH: str(len(content))}
            net.add_md5_header(response_headers, 'beef')
            session.put.return_value = Mock(headers=response_headers)

            with pytest.raises(exceptions.TransferException):
                client.upload_file(url=target_url, src=src.name)

        finally:
            ws.MAX_MD5_COMPUTE_SIZE = orig_max_md5_compute_size

        # empty file
        empty_file = tempfile.NamedTemporaryFile()
        with open(empty_file.name, 'w') as f:
            f.write('')
        with pytest.raises(ValueError):
            client.upload_file(url=target_url,
                               src=empty_file.name)

    @patch('cadcutils.net.ws.WsCapabilities')
    def test_upload_file_put_txn_append(self, caps_mock):
        anon_subject = auth.Subject()
        target_url = 'https://someurl'
        cm = Mock()
        cm.get_access_url.return_value = "http://host/availability"
        caps_mock.return_value = cm
        response = Mock()
        response.status_code = requests.codes.ok
        response.headers = {}
        session = Mock()
        session.put.return_value = response
        client = ws.BaseDataClient(resource_id='ivo://cadc.nrc.ca/resourceid',
                                   subject=anon_subject, agent='TestApp')
        client._get_session = Mock(return_value=session)

        orig_max_md5_compute_size = ws.MAX_MD5_COMPUTE_SIZE
        orig_max_file_segment_size = ws.FILE_SEGMENT_THRESHOLD
        try:
            ws.MAX_MD5_COMPUTE_SIZE = 5  # force transaction
            ws.FILE_SEGMENT_THRESHOLD = 10  # force segments
            segments = [b'segment1', b'segment2', b'end3']
            src = tempfile.NamedTemporaryFile()
            start_txn_headers = {ws.PUT_TXN_ID: '123',
                                 ws.PUT_TXN_MIN_SEGMENT: '1',
                                 ws.PUT_TXN_MAX_SEGMENT: len(segments[0])}

            def _create_put_responses():
                # this function creates the content of the file according
                # to the segments and also generates the responses to
                # PUT commands with corresponding headers
                md5 = hashlib.md5()
                seg_md5s = []
                with open(src.name, 'wb') as f:
                    for seg in segments:
                        f.write(seg)
                        md5.update(seg)
                        seg_md5s.append(md5.hexdigest())
                commit_txn_headers = {ws.HTTP_LENGTH: '0'}
                net.add_md5_header(commit_txn_headers, seg_md5s[-1])
                responses = [Mock(headers=start_txn_headers)]
                for seg_md5 in seg_md5s:
                    segment_txn_headers = {ws.PUT_TXN_ID: '123',
                                           ws.HTTP_LENGTH: '0'}
                    net.add_md5_header(segment_txn_headers, seg_md5)
                    responses.append(
                        Mock(headers=segment_txn_headers))

                responses.append(Mock(headers=commit_txn_headers))
                return responses

            put_responses = _create_put_responses()
            file_size = os.stat(src.name).st_size

            def put_mock(url, data=None, **kwargs):
                # this is a "semi" mock of the PUT function. The PUT request
                # is mocked but this function consumes from the file (data)
                # when available
                headers = kwargs['headers']
                assert url == target_url
                if put_mock.put_num:
                    # except the first PUT that starts transaction,
                    # all the other ones need to include the trans id
                    assert headers[ws.PUT_TXN_ID] == '123'
                    if put_mock.put_num in [1, 2, 3]:
                        if put_mock.exception is not None and \
                                put_mock.exception == put_mock.put_num:
                            # raise a Transfer exception for the segment
                            session.head.return_value = put_responses[
                                put_mock.exception]
                            session.post.return_value = \
                                put_responses[put_mock.exception - 1]
                            put_mock.exception = None
                            raise exceptions.TransferException('Test')
                        if put_mock.wrong_md5 is not None and \
                                (put_mock.put_num == put_mock.wrong_md5):
                            # return the wrong md5 for the segment
                            # set the response for the revert POST to be
                            # similar to the previous (successful) PUT
                            session.post.return_value = \
                                put_responses[put_mock.wrong_md5 - 1]
                            if put_mock.wrong_md5 == 1:
                                # first segment - return start txn response
                                rsp = put_responses[0]
                            else:
                                # return a mismatch md5 response
                                tmp_hd = {ws.PUT_TXN_ID: '123',
                                          ws.HTTP_LENGTH: '0'}
                                net.add_md5_header(tmp_hd, 'beef')
                                rsp = Mock(headers=tmp_hd)
                            put_mock.wrong_md5 = None
                            return rsp

                        # check length of segments
                        assert headers[ws.HTTP_LENGTH] == \
                            str(len(segments[put_mock.put_num-1]))
                        assert headers[CONTENT_TYPE] == TEXT_TYPE
                    else:
                        assert headers[ws.PUT_TXN_OP] == ws.PUT_TXN_COMMIT
                        assert headers[ws.HTTP_LENGTH] == '0'
                        assert headers[CONTENT_TYPE] == TEXT_TYPE
                else:
                    assert headers[ws.PUT_TXN_OP] == ws.PUT_TXN_START
                    assert headers[ws.HTTP_LENGTH] == '0'
                    assert headers[ws.PUT_TXN_TOTAL_LENGTH] == str(file_size)
                    assert headers[CONTENT_TYPE] == TEXT_TYPE
                if data:
                    data.read(100)
                rsp = put_responses[put_mock.put_num]
                put_mock.put_num += 1
                return rsp
            put_mock.exception = None
            put_mock.put_num = 0
            put_mock.wrong_md5 = None
            session.put = put_mock
            # add caller headers and test they are passed through
            caller_header = {CONTENT_TYPE: TEXT_TYPE}
            client.upload_file(url=target_url, src=src.name,
                               headers=caller_header)
            # check all puts were called
            assert len(put_responses) == put_mock.put_num

            # redo the tests but have the file size multiple of its segments
            # reset the put_mock "mock" function first

            put_mock.exception = None
            put_mock.put_num = 0
            put_mock.wrong_md5 = None
            segments = [b'segment1', b'segment2', b'endsize3']
            put_responses = _create_put_responses()
            file_size = os.stat(src.name).st_size

            client.upload_file(url=target_url, src=src.name,
                               headers=caller_header)
            # check all puts were called
            assert len(put_responses) == put_mock.put_num

            # redo the test but have an exception thrown in PUT for segment 2
            put_mock.exception = 2
            put_mock.put_num = 0
            client.upload_file(url=target_url, src=src.name,
                               headers=caller_header)
            # check all puts were called
            assert len(put_responses) == put_mock.put_num

            # redo the test but have a md5 mismatch for segment 1
            put_mock.wrong_md5 = 1
            put_mock.put_num = 0
            client.upload_file(url=target_url, src=src.name,
                               headers=caller_header)
            # check all puts were called
            assert len(put_responses) == put_mock.put_num

            # repeat the test but have a md5 mismatch for segment 2
            put_mock.wrong_md5 = 2
            put_mock.put_num = 0
            client.upload_file(url=target_url, src=src.name,
                               headers=caller_header)
            # check all puts were called
            assert len(put_responses) == put_mock.put_num

            # permanent Transfer error
            session.put = Mock(Mock(headers=start_txn_headers),
                               side_effect=[exceptions.TransferException] * 3)
            with pytest.raises(exceptions.TransferException):
                client.upload_file(url=target_url, src=src.name)

            # permanent Transfer error including in HEAD
            session.put = Mock(side_effect=[Mock(headers=start_txn_headers),
                                            exceptions.TransferException])
            head_exception = exceptions.UnexpectedException
            session.head = Mock(side_effect=head_exception)
            with pytest.raises(head_exception):
                client.upload_file(url=target_url, src=src.name)
        finally:
            ws.MAX_MD5_COMPUTE_SIZE = orig_max_md5_compute_size
            ws.PUT_TXN_MAX_SEGMENT = orig_max_file_segment_size

    def test_get_segment_size(self):
        get_segment = ws.BaseDataClient._get_segment_size  # shortcut
        # file size > preferred segment size
        assert ws.PREFERRED_SEGMENT_SIZE == \
            get_segment(ws.PREFERRED_SEGMENT_SIZE+1, None, None)
        # file size < preferred segment size
        assert ws.PREFERRED_SEGMENT_SIZE - 1 == \
            get_segment(ws.PREFERRED_SEGMENT_SIZE-1, None, None)
        # minimum size > file_size and preferred size
        assert ws.PREFERRED_SEGMENT_SIZE + 2 == \
            get_segment(ws.PREFERRED_SEGMENT_SIZE,
                        ws.PREFERRED_SEGMENT_SIZE+2, None)
        # max size < file_size and preferred size
        assert ws.PREFERRED_SEGMENT_SIZE - 1 == \
            get_segment(ws.PREFERRED_SEGMENT_SIZE,
                        None, ws.PREFERRED_SEGMENT_SIZE-1)


class TestRetrySession(unittest.TestCase):
    """ Class for testing retry session """

    @patch('time.sleep')
    @patch('cadcutils.net.ws.requests.Session.send')
    @patch('cadcutils.net.ws.requests.Session.merge_environment_settings',
           Mock(return_value={}))
    def test_retry(self, send_mock, time_mock):
        request = Mock()
        send_mock.return_value = Mock()
        rs = ws.RetrySession(False)
        rs.send(request)
        send_mock.assert_called_with(request, timeout=120)

        # retry to user defined timeout
        send_mock.return_value = Mock()
        rs = ws.RetrySession(False)
        rs.send(request, timeout=77)
        send_mock.assert_called_with(request, timeout=77)

        # mock delays for connect timeout
        send_mock.reset_mock()
        rs = ws.RetrySession()
        # connection error that triggers retries
        cte = requests.exceptions.ConnectTimeout()
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [cte, response]
        rs.send(request)
        time_mock.assert_called_with(DEFAULT_RETRY_DELAY)

        # mock delays for read timeout
        send_mock.reset_mock()
        rs = ws.RetrySession()
        # connection error that triggers retries
        rte = requests.exceptions.ReadTimeout()
        send_mock.side_effect = [rte]
        with self.assertRaises(exceptions.TransferException):
            rs.send(request)

        # mock Connection errors
        send_mock.reset_mock()
        rs = ws.RetrySession()
        # connection error that triggers retries
        ce = requests.exceptions.ConnectionError('Some error')
        send_mock.side_effect = [ce]
        with self.assertRaises(exceptions.HttpException):
            rs.send(request)

        # mock reset by peer error
        # mock Connection errors
        send_mock.reset_mock()
        rs = ws.RetrySession()
        # connection error that triggers retries
        ce = requests.exceptions.ConnectionError('Connection reset by peer')
        send_mock.side_effect = [ce]
        with self.assertRaises(exceptions.TransferException):
            rs.send(request)

        # two connection error delay = DEFAULT_RETRY_DELAY
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession()
        # connection error that triggers retries
        ce = requests.exceptions.ConnectTimeout()
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [ce, ce, response]  # two connection errors
        rs.send(request)
        calls = [call(DEFAULT_RETRY_DELAY), call(DEFAULT_RETRY_DELAY * 2)]
        time_mock.assert_has_calls(calls)

        # set the start retry to a large number and see how it is capped
        # to MAX_RETRY_DELAY
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession(start_delay=MAX_RETRY_DELAY / 2 + 1)
        # connection error that triggers retries
        ce = requests.exceptions.ConnectTimeout()
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [ce, ce, response]  # two connection errors
        rs.send(request)
        calls = (call(MAX_RETRY_DELAY / 2 + 1), call(MAX_RETRY_DELAY))
        time_mock.assert_has_calls(calls)

        # return the error all the time
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession(start_delay=MAX_RETRY_DELAY / 2 + 1)
        # connection error that triggers retries
        ce = requests.exceptions.ConnectTimeout()
        # make sure the mock returns more errors than the maximum number
        # of retries allowed
        http_errors = []
        i = 0
        while i <= MAX_NUM_RETRIES:
            http_errors.append(ce)
            i += 1
        send_mock.side_effect = http_errors
        with self.assertRaises(exceptions.HttpException):
            rs.send(request)

        # return HttpError 503 with Retry-After
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession()
        server_delay = 5
        # connection error that triggers retries
        he = requests.exceptions.HTTPError()
        he.response = requests.Response()
        he.response.status_code = requests.codes.unavailable
        he.response.headers[SERVICE_RETRY] = server_delay
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [he, response]
        rs.send(request)
        calls = [call(server_delay)]
        time_mock.assert_has_calls(calls)

        # return HttpError 503 with Retry-After with an invalid value
        send_mock.reset_mock()
        time_mock.reset_mock()
        start_delay = 66
        rs = ws.RetrySession(start_delay=start_delay)
        server_delay = 'notnumber'
        # connection error that triggers retries
        he = requests.exceptions.HTTPError()
        he.response = requests.Response()
        he.response.status_code = requests.codes.unavailable
        he.response.headers[SERVICE_RETRY] = server_delay
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [he, response]
        rs.send(request)
        calls = [call(start_delay)]  # uses the default delay
        time_mock.assert_has_calls(calls)

        # return HttpError 503 with no Retry-After
        send_mock.reset_mock()
        time_mock.reset_mock()
        start_delay = 66
        rs = ws.RetrySession(start_delay=start_delay)
        he = requests.exceptions.HTTPError()
        he.response = requests.Response()
        he.response.status_code = requests.codes.unavailable
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [he, response]
        rs.send(request)
        calls = [call(start_delay)]  # uses the default delay
        time_mock.assert_has_calls(calls)

        # tests non-transient errors
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession()
        he = requests.exceptions.HTTPError()
        he.response = requests.Response()
        he.response.status_code = requests.codes.internal_server_error
        send_mock.side_effect = he
        with self.assertRaises(exceptions.HttpException):
            rs.send(request)

    def test_config_file_location(self):
        # test the location of the config file when the host is specified
        myhost = 'myhost.ca'
        service = 'myservice'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        client = Mock(resource_id=resource_id)
        with patch('cadcutils.net.ws.os.makedirs') as makedirs_mock:
            wscap = ws.WsCapabilities(client, host=myhost)
        expected_location = os.path.join(ws.CACHE_LOCATION, 'alt-domains',
                                         myhost)
        makedirs_mock.check_called_with(expected_location)
        self.assertEqual(os.path.join(expected_location, 'resource-caps'),
                         wscap.reg_file)
        self.assertEqual(
            os.path.join(expected_location, urlparse(resource_id).netloc,
                         '.{}'.format(service)),
            wscap.caps_file)

    def test_misc(self):
        """
        Tests miscellaneous functions
        """
        service_url = 'http://somehost.com/service'
        with patch('cadcutils.net.ws.WsCapabilities') as caps_mock:
            caps_mock.return_value.get_service_host.return_value =\
                'somehost.com'
            caps_mock.return_value.get_access_url.return_value = service_url
            client = ws.BaseWsClient("someresourceID", auth.Subject(),
                                     'TestApp')
            self.assertEqual('{}'.format(service_url),
                             client._get_url(('myfeature', None)))
            caps_mock.return_value.get_access_url.assert_called_with(
                'myfeature')
            self.assertEqual('{}'.format(service_url),
                             client._get_url(('myfeature', '')))

        test_host = 'testhost.com'
        with patch('cadcutils.net.ws.os.makedirs'):
            client = ws.BaseWsClient("someresourceID", auth.Subject(),
                                     'TestApp', host=test_host)
        self.assertEqual(test_host, client.host)

        # test with resource as url
        with patch('cadcutils.net.ws.WsCapabilities') as caps_mock:
            caps_mock.return_value.get_service_host.return_value =\
                'somehost.com'
            cm = Mock()
            cm.get_access_url.return_value = "http://host/availability"
            caps_mock.return_value = cm
            client = ws.BaseWsClient("someresourceID", auth.Subject(),
                                     'TestApp')
            resource_url = 'http://someurl.com/path/'
            self.assertEqual(resource_url, client._get_url(resource_url))
            # repeat with overriden host name
            client = ws.BaseWsClient("someresourceID", auth.Subject(),
                                     'TestApp', host=test_host)
            # same result
            self.assertEqual(resource_url, client._get_url(resource_url))

            # test exceptions with different status in the response
            session = ws.RetrySession()
            response = requests.Response()
            response.status_code = requests.codes.not_found
            with self.assertRaises(exceptions.NotFoundException):
                session.check_status(response)
            response.status_code = requests.codes.unauthorized
            with self.assertRaises(exceptions.UnauthorizedException):
                session.check_status(response)
            response.status_code = requests.codes.forbidden
            with self.assertRaises(exceptions.ForbiddenException):
                session.check_status(response)
            response.status_code = requests.codes.bad_request
            with self.assertRaises(exceptions.BadRequestException):
                session.check_status(response)
            response.status_code = requests.codes.conflict
            with self.assertRaises(exceptions.AlreadyExistsException):
                session.check_status(response)
            response.status_code = requests.codes.internal_server_error
            with self.assertRaises(exceptions.InternalServerException):
                session.check_status(response)
            response.status_code = requests.codes.unavailable
            with self.assertRaises(requests.HTTPError):
                session.check_status(response)
            response.status_code = requests.codes.not_extended
            with self.assertRaises(exceptions.UnexpectedException):
                session.check_status(response)

    @patch('time.sleep')
    @patch('cadcutils.net.ws.requests.Session.send')
    def test_idempotent(self, send_mock, time_mock):
        # mock delays for connect timeout
        send_mock.reset_mock()
        rs = ws.RetrySession()
        # GET, PUT, DELETE, HEAD are idempotent operations so re-tries on
        # connection timeouts happen automatically
        cte = requests.exceptions.ConnectTimeout()
        response = requests.Response()
        response.status_code = requests.codes.ok
        # GET
        send_mock.side_effect = [cte, response]
        rs.get('https://someurl')
        time_mock.assert_called_with(DEFAULT_RETRY_DELAY)
        # PUT
        send_mock.reset_mock()
        time_mock.reset_mock()
        send_mock.side_effect = [cte, response]
        rs.put('https://someurl')
        time_mock.assert_called_with(DEFAULT_RETRY_DELAY)
        # DELETE
        send_mock.reset_mock()
        time_mock.reset_mock()
        send_mock.side_effect = [cte, response]
        rs.delete('https://someurl')
        time_mock.assert_called_with(DEFAULT_RETRY_DELAY)
        # HEAD
        send_mock.reset_mock()
        time_mock.reset_mock()
        send_mock.side_effect = [cte, response]
        rs.head('https://someurl')
        time_mock.assert_called_with(DEFAULT_RETRY_DELAY)

        # POST is not idempotent by default so re-try does not happen
        send_mock.reset_mock()
        time_mock.reset_mock()
        send_mock.side_effect = [cte, response]
        with pytest.raises(requests.exceptions.ConnectTimeout):
            rs.post('https://someurl')

        # POST with service unavailable response. Note the different exception
        # when exception is generated from response
        unavailable = requests.Response()
        unavailable.status_code = requests.codes.unavailable
        send_mock.side_effect = [unavailable]
        with pytest.raises(exceptions.UnexpectedException):
            rs.post('https://someurl')

        # Create session with idempotent POSTs to enable retries
        send_mock.reset_mock()
        time_mock.reset_mock()
        send_mock.side_effect = [cte, response]
        rs = ws.RetrySession(idempotent_posts=True)
        rs.post('https://someurl')
        time_mock.assert_called_with(DEFAULT_RETRY_DELAY)


capabilities__content = \
    """
    <vosi:capabilities
    xmlns:vosi="http://www.ivoa.net/xml/VOSICapabilities/v1.0"
    xmlns:vs="http://www.ivoa.net/xml/VODataService/v1.1"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        <capability standardID="ivo://ivoa.net/std/VOSI#capabilities">
            <interface xsi:type="vs:ParamHTTP" role="std">
                <accessURL use="full">
                    http://WS_URL/capabilities
                </accessURL>
            </interface>
        </capability>
        <capability standardID="ivo://ivoa.net/std/VOSI#availability">
            <interface xsi:type="vs:ParamHTTP" role="std">
                <accessURL use="full">
                    http://WS_URL/availability
                </accessURL>
            </interface>
        </capability>
        <capability standardID="vos://cadc.nrc.ca~service/CADC/mystnd01">
            <interface xsi:type="vs:ParamHTTP" role="std" version="1.0">
                <accessURL use="base">
                    http://WS_URL/pub
                </accessURL>
            </interface>
            <interface xsi:type="vs:ParamHTTP" role="std" version="1.0">
                <accessURL use="base">
                    http://WS_URL/auth
                </accessURL>
                <securityMethod
                standardID=
                "ivo://ivoa.net/sso#BasicAA"/>
            </interface>
            <interface xsi:type="vs:ParamHTTP" role="std" version="1.0">
                <accessURL use="base">
                    https://WS_URL
                </accessURL>
                <securityMethod
                standardID="ivo://ivoa.net/sso#tls-with-certificate"/>
            </interface>
        </capability>
    </vosi:capabilities>
    """


class TestWsCapabilities(unittest.TestCase):
    """Class for testing the webservie client"""

    def test_get_content(self):
        """
        Sometimes servers return empty capabilities documents and
        the client is expected to re-use the cached document
        :return:
        """
        ws_client = Mock(resource_id='SOME_RESOURCE')
        caps = ws.WsCapabilities(ws_client)
        now = time.time()
        resource_file = tempfile.NamedTemporaryFile()
        open(resource_file.name, 'w').write('OLD CONTENT')
        with patch('cadcutils.net.ws.requests.Session.get') as mock_get:
            mock_get.return_value = Mock(text='')
            assert 'OLD CONTENT' == \
                   caps._get_content(resource_file.name, 'some/url',
                                     now-ws.CACHE_REFRESH_INTERVAL-1)

    @patch('cadcutils.net.ws.os.path.getmtime')
    @patch('cadcutils.net.ws.open', mock=mock_open())
    @patch('cadcutils.net.ws.requests.Session.get')
    def test_get_reg(self, get_mock, file_mock, file_modtime_mock):
        """
        Tests the registry part of WsCapabilities
        """
        # test when registry information is read from the server
        # (cache is outdated)
        service = 'myservice'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        resource_cap_url = 'http://www.canfar.net/myservice'
        cadcreg_content = ('#test content\n {} = {} \n'
                           'ivo://some.provider/service = '
                           'http://providerurl.test/service'). \
            format(resource_id, resource_cap_url)
        response = Mock(text=cadcreg_content)
        get_mock.return_value = response
        # set the modified time of the cache file to 0 to make sure the info
        # is retrieved from server
        file_modtime_mock.return_value = 0
        # test anonymous access
        fh_mock = Mock()
        file_mock.write = fh_mock
        client = Mock(resource_id=resource_id)
        caps = ws.WsCapabilities(client)
        self.assertEqual(os.path.join(ws.CACHE_LOCATION, ws.REGISTRY_FILE),
                         caps.reg_file)
        self.assertEqual(
            os.path.join(ws.CACHE_LOCATION,
                         'canfar.phys.uvic.ca', '.{}'.format(service)),
            caps.caps_file)
        self.assertEqual(resource_cap_url, caps._get_capability_url())
        file_mock.assert_called_once_with(
            os.path.join(ws.CACHE_LOCATION, ws.REGISTRY_FILE), 'w')
        # TODO not sure why need to access write this way
        file_mock().__enter__.return_value.write.assert_called_once_with(
            cadcreg_content)

        # test when registry information is retrieved from the cache file
        get_mock.reset_mock()
        get_mock.return_value = None
        file_modtime_mock.reset_mock()
        file_mock.reset_mock()
        resource_cap_url2 = 'http://www.canfar.net/myservice2'
        cache_content2 = ('#test content\n {} = {} \n'
                          'ivo://some.provider/service = '
                          'http://providerurl.test/service'). \
            format(resource_id, resource_cap_url2)
        file_modtime_mock.return_value = time.time()
        file_mock().__enter__.return_value.read.return_value = cache_content2
        caps = ws.WsCapabilities(client)
        self.assertEqual(resource_cap_url2, caps._get_capability_url())

        # test when registry information is outdated but there are
        # errors retrieving it from the CADC registry
        # so in the end go back and use the cache version
        file_modtime_mock.reset_mock()
        file_mock.reset_mock()
        file_modtime_mock.return_value = 0
        file_mock().__enter__.return_value.read.return_value = cache_content2
        get_mock.side_effect = [exceptions.HttpException()]
        client.get.side_effect = [exceptions.HttpException]
        caps = ws.WsCapabilities(client)
        with patch('os.path.exists', Mock()):
            self.assertEqual(resource_cap_url2, caps._get_capability_url())

    @patch('cadcutils.net.ws.os.path.getmtime')
    @patch('cadcutils.net.ws.requests.Session.get')
    def test_get_caps(self, get_mock, file_modtime_mock):
        """
        Tests the capabilities part of WsCapabilities
        """
        # test when registry information is read from the server
        # (cache is outdated)
        service = 'myservice'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        resource_cap_url = 'www.canfar.net/myservice'
        # set the modified time of the cache file to 0 to make sure the
        # info is retrieved from server
        file_modtime_mock.return_value = 0
        # test anonymous access
        expected_content = capabilities__content.replace('WS_URL',
                                                         resource_cap_url)
        response = Mock(text=expected_content)
        get_mock.return_value = response
        caps = ws.WsCapabilities(Mock(resource_id=resource_id,
                                      subject=auth.Subject()))

        # mock _get_capability_url to return some url without attempting
        # to access the server
        def get_url():
            return 'http://some.url/capabilities'

        caps._get_capability_url = get_url
        # remove the cached file if exists
        if os.path.isfile(caps.caps_file):
            os.remove(caps.caps_file)
        caps.caps_urls[service] = '{}/capabilities'.format(resource_cap_url)
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('http://{}/pub'.format(resource_cap_url),
                         caps.get_access_url(
                             'vos://cadc.nrc.ca~service/CADC/mystnd01'))
        actual_content = open(caps.caps_file, 'r').read()
        self.assertEqual(expected_content, actual_content)

        # mock _get_capability_url to return a subservice
        service = 'myservice/mysubservice'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        resource_cap_url = 'www.canfar.net/myservice/mysubservice'
        # set the modified time of the cache file to 0 to make sure the
        # info is retrieved from server
        file_modtime_mock.return_value = 0
        expected_content = capabilities__content.replace('WS_URL',
                                                         resource_cap_url)
        # test anonymous access
        response = Mock(text=expected_content)
        get_mock.return_value = response
        caps = ws.WsCapabilities(Mock(resource_id=resource_id,
                                      subject=auth.Subject()))
        # remove the cached file if exists
        if os.path.isfile(caps.caps_file):
            os.remove(caps.caps_file)
        caps._get_capability_url = get_url
        caps.caps_urls[service] = '{}/capabilities'.format(resource_cap_url)
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('http://{}/pub'.format(resource_cap_url),
                         caps.get_access_url(
                             'vos://cadc.nrc.ca~service/CADC/mystnd01'))
        actual_content = open(caps.caps_file, 'r').read()
        self.assertEqual(expected_content, actual_content)

        # repeat for basic auth subject. Mock the netrc library to
        # prevent a lookup for $HOME/.netrc
        with patch('cadcutils.net.auth.netrclib'):
            client = Mock(resource_id=resource_id,
                          subject=auth.Subject(netrc=True))
        client.get.return_value = response
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # capabilities works even if it has only one anonymous interface
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#capabilities'))
        # same for availability
        self.assertEqual('http://{}/availability'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#availability'))

        # repeat for https
        with patch('os.path.isfile'):
            client = Mock(resource_id=resource_id,
                          subject=auth.Subject(certificate='somecert.pem'))
        client.get.return_value = response
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # capabilities works even if it has only one anonymous interface
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#capabilities'))
        # same for availability
        self.assertEqual('http://{}/availability'.format(resource_cap_url),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('https://{}'.format(resource_cap_url),
                         caps.get_access_url(
                             'vos://cadc.nrc.ca~service/CADC/mystnd01'))

        # test when capabilities information is retrieved from the cache file
        get_mock.reset_mock()
        get_mock.return_value = None
        file_modtime_mock.reset_mock()
        service = 'myservice2'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        resource_cap_url2 = 'canfar.phys.uvic.ca/myservice2'
        expected_content = capabilities__content.replace('WS_URL',
                                                         resource_cap_url2)
        file_modtime_mock.return_value = time.time()
        client = Mock(resource_id=resource_cap_url2, subject=auth.Subject())
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        caps.caps_urls[service] = '{}/capabilities'.format(resource_cap_url2)
        # manually write the content
        with open(caps.caps_file, 'w') as f:
            f.write(expected_content)
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url2),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url2),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('http://{}/pub'.format(resource_cap_url2),
                         caps.get_access_url(
                             'vos://cadc.nrc.ca~service/CADC/mystnd01'))
        actual_content = open(caps.caps_file, 'r').read()
        self.assertEqual(expected_content, actual_content)

        # repeat for basic auth subject. Mock the netrc library to prevent a
        # lookup for $HOME/.netrc
        with patch('cadcutils.net.auth.netrclib'):
            client = Mock(resource_id=resource_id,
                          subject=auth.Subject(netrc=True))
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # does not work with user-password of the subject set above
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url2),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url2),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('http://{}/auth'.format(resource_cap_url2),
                         caps.get_access_url(
                             'vos://cadc.nrc.ca~service/CADC/mystnd01'))

        # repeat for https
        with patch('os.path.isfile'):
            client = Mock(resource_id=resource_id,
                          subject=auth.Subject(certificate='somecert.pem'))
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # does not work with user-password of the subject set above
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url2),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url2),
                         caps.get_access_url(
                             'ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('https://{}'.format(resource_cap_url2),
                         caps.get_access_url(
                             'vos://cadc.nrc.ca~service/CADC/mystnd01'))


class TestWsOutsideCalls(unittest.TestCase):
    """ Class to test Ws with calls to outside sites"""

    @patch('time.sleep')
    def testCalls(self, time_mock):
        client = ws.BaseWsClient('https://httpbin.org', net.Subject(), 'FOO')
        response = client.get('https://httpbin.org')
        self.assertEqual(response.status_code, requests.codes.ok)

        with self.assertRaises(exceptions.InternalServerException):
            client.get('https://httpbin.org/status/500')

        time_mock.reset_mock()
        with self.assertRaises(exceptions.HttpException):
            client.get('https://httpbin.org/status/503')

        calls = [call(DEFAULT_RETRY_DELAY),
                 call(min(DEFAULT_RETRY_DELAY*2, MAX_RETRY_DELAY)),
                 call(min(DEFAULT_RETRY_DELAY * 4, MAX_RETRY_DELAY)),
                 call(min(DEFAULT_RETRY_DELAY * 8, MAX_RETRY_DELAY)),
                 call(min(DEFAULT_RETRY_DELAY * 16, MAX_RETRY_DELAY)),
                 call(min(DEFAULT_RETRY_DELAY * 32, MAX_RETRY_DELAY))]

        time_mock.assert_has_calls(calls)


def test_resolve_name():
    client = ws.BaseDataClient('https://httpbin.org', net.Subject(), 'FOO')
    dest = NamedTemporaryFile()
    dest_dir = os.path.dirname(dest.name)
    file_name = os.path.basename(dest.name)
    tmp_file = os.path.join(dest_dir, '{}-beef.part'.format(file_name))
    # file name provided
    final_dest, temp_dest = client._resolve_destination_file(
        dest.name, 'beef', None)
    assert dest.name == final_dest
    assert tmp_file == temp_dest

    # directory provided
    dest_dir = os.path.dirname(dest.name)
    final_dest, temp_dest = client._resolve_destination_file(
        dest_dir, 'beef', os.path.basename(dest.name))
    assert dest.name == final_dest
    assert tmp_file == temp_dest

    # no md5 - no temporary file/resume
    final_dest, temp_dest = client._resolve_destination_file(
        dest_dir, None, os.path.basename(dest.name))
    assert dest.name == final_dest == temp_dest

    with pytest.raises(ValueError):
        client._resolve_destination_file(None, 'beef', None)


def test_download_file_method():
    client = ws.BaseDataClient('https://httpbin.org', net.Subject(), 'FOO')
    # mock the get
    response = Mock()
    file_name = 'filename.jpg'
    md5 = 'ab56b4d92b40713acc5af89985d4b786'
    response.status_code = requests.codes.ok
    response.headers = \
        {'digest': 'md5=YWI1NmI0ZDkyYjQwNzEzYWNjNWFmODk5ODVkNGI3ODY=',
         'Content-Length': '5',
         'content-disposition': 'attachment; filename={}'.format(file_name)}
    response.raw.read.side_effect = [b'abc', b'de']
    client.get = Mock(return_value=response)
    temp_dir = TemporaryDirectory()
    # proxy _save_bytes through a mock to check when it's called
    client._save_bytes = Mock(side_effect=client._save_bytes)
    client.download_file('https://dataservice', temp_dir.name)
    dest, temp_dest = client._resolve_destination_file(
        temp_dir.name, src_md5=md5, default_file_name=file_name)
    assert os.path.isfile(dest)
    assert not os.path.isfile(temp_dest)
    assert client._save_bytes.called
    client.get.assert_called_once_with('https://dataservice', stream=True)

    # calling it the second time does not cause another get
    client.get.reset_mock()
    client._save_bytes.reset_mock()
    client.get = Mock(return_value=response)
    client.download_file('https://dataservice', temp_dir.name)
    assert not client._save_bytes.called
    client.get.assert_called_once_with('https://dataservice', stream=True)

    # make temporary file larger (BUG case). Operation should succeed
    client.get.reset_mock()
    os.rename(dest, temp_dest)
    open(temp_dest, 'ab').write(b'ghi')
    assert 5 < os.stat(temp_dest).st_size
    response.raw.read.side_effect = [b'abc', b'de']
    client.download_file('https://dataservice', temp_dir.name)
    dest, temp_dest = client._resolve_destination_file(
        temp_dir.name, src_md5=md5, default_file_name=file_name)
    assert os.path.isfile(dest)
    assert not os.path.isfile(temp_dest)
    assert client._save_bytes.called
    client.get.assert_called_once_with('https://dataservice', stream=True)

    # calling it when incomplete temporary file exits
    # truncate the last 3 bytes but the service does not support
    # ranges
    client.get.reset_mock()
    os.rename(dest, temp_dest)
    with open(temp_dest, 'r+') as f:
        f.seek(0, os.SEEK_END)
        f.seek(f.tell() - 3, os.SEEK_SET)
        f.truncate()
    response.raw.read.side_effect = [b'abc', b'de']
    client.get = Mock(return_value=response)
    client.download_file('https://dataservice', temp_dir.name)
    assert os.path.isfile(dest)
    assert not os.path.isfile(temp_dest)
    assert client._save_bytes.called
    client.get.assert_called_once_with('https://dataservice', stream=True)

    # repeat the test when the service supports ranges
    client.get.reset_mock()
    os.rename(dest, temp_dest)
    with open(temp_dest, 'r+') as f:
        f.seek(0, os.SEEK_END)
        f.seek(f.tell() - 3, os.SEEK_SET)
        f.truncate()
    response.status_code = requests.codes.partial_content
    response.headers['Accept-Ranges'] = 'bytes '
    response.raw.read.side_effect = [b'cde']
    client.get = Mock(return_value=response)
    client.download_file('https://dataservice', temp_dir.name)
    assert os.path.isfile(dest)
    assert not os.path.isfile(temp_dest)
    assert client._save_bytes.called
    assert 2 == client.get.call_count
    # second get call with Range header expected
    assert [call('https://dataservice', stream=True),
            call('https://dataservice', stream=True,
                 headers={'Range': 'bytes=2-'})] in client.get.mock_calls


def test_save_bytes():
    client = ws.BaseDataClient('https://httpbin.org', net.Subject(), 'FOO')
    dest = NamedTemporaryFile()
    outer = {'bytes_count': 0}

    def count_bytes(bytes):
        outer['bytes_count'] += len(bytes)

    response = Mock()
    response.raw.read.side_effect = [b'abc', b'de']
    response.headers = \
        {'digest': 'md5=YWI1NmI0ZDkyYjQwNzEzYWNjNWFmODk5ODVkNGI3ODY='}
    client._save_bytes(response=response, src_length=5, dest_file=dest.name,
                       process_bytes=count_bytes)
    assert 5 == outer['bytes_count']
    assert 5 == os.stat(dest.name).st_size
    assert b'abcde' == open(dest.name, 'rb').read()

    # repeat the test but make szie of source and destination mismatch
    dest = NamedTemporaryFile()
    response = Mock()
    response.raw.read.side_effect = [b'aaa', b'aa']  # different content
    response.headers = \
        {'digest': 'md5=YWI1NmI0ZDkyYjQwNzEzYWNjNWFmODk5ODVkNGI3ODY='}
    with pytest.raises(exceptions.TransferException):
        client._save_bytes(response=response, src_length=7,
                           dest_file=dest.name,
                           process_bytes=count_bytes)

    # repeat the test but make md5s of source and destination mismatch
    dest = NamedTemporaryFile()
    response = Mock()
    response.raw.read.side_effect = [b'aaa', b'aa']  # different content
    response.headers = \
        {'digest': 'md5=YWI1NmI0ZDkyYjQwNzEzYWNjNWFmODk5ODVkNGI3ODY='}
    with pytest.raises(exceptions.TransferException):
        client._save_bytes(response=response, src_length=5,
                           dest_file=dest.name,
                           process_bytes=count_bytes)



# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2021.                            (c) 2021.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

"""
Module that contains functionality related to CADC groups.
"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import sys
import traceback
import logging
from six.moves.urllib.parse import urlparse, urlencode


from cadcutils import version  # TODO should it be application version instead?
from cadcutils import util, exceptions
from . import BaseWsClient, Subject
from .auth import SECURITY_METHODS_IDS, CookieInfo
from . import Role, Group, Identity
from .group_xml import GroupWriter, GroupReader, GroupsReader


CADC_LOGIN_CAPABILITY = 'ivo://ivoa.net/std/UMS#login-0.1'
CADC_SSO_COOKIE_NAME = 'CADC_SSO'
CADC_REALMS = ['.canfar.net', '.cadc-ccda.hia-iha.nrc-cnrc.gc.ca',
               '.cadc.dao.nrc.ca']

# ID of the default catalog Web service
CADC_AC_SERVICE_ID = 'ivo://cadc.nrc.ca/gms'
DEFAULT_SERVICE_ID = 'ivo://cadc.nrc.ca/gms'

SEARCH_CAPABILITY_ID = 'ivo://ivoa.net/std/GMS#search-1.0'

# spec extensions relative to root of GMS service URL
GROUPS_PATH = 'groups'
GROUP_MEMBER_PATH = 'groupMembers'
USER_MEMBER_PATH = 'userMembers'

APP_NAME = 'cadc-group'

__all__ = ['GroupsClient']

logger = util.get_logger(__name__)

cadcgms_agent = 'cadc-gms-client/{}'.format(version.version)


class GroupsClient():
    """Class for interacting with the access control web service"""

    def __init__(self, subject, resource_id=DEFAULT_SERVICE_ID,
                 host=None, agent=None, insecure=False):
        """
        Instance of a GroupsClient
        :param subject: the subject performing the action
        :type cadcutils.auth.Subject
        :param resource_id: the resource ID of the service (service ID)
        :param host: Host for the GMS service (use for testing mainly)
        :param agent: The name of the agent (to be used in server logging)
        :param insecure Allow insecure server connections over SSL (testing)
        """
        self.resource_id = resource_id
        self.host = host

        self._subject = subject
        if agent is None:
            self.agent = cadcgms_agent
        else:
            self.agent = agent

        # for the CADC TAP services, BasicAA is not supported anymore, so
        # we need to login and get a cookie when the subject uses
        # user/passwd
        if resource_id.startswith('ivo://cadc.nrc.ca') and\
           SECURITY_METHODS_IDS['basic'] in subject.get_security_methods():
            login = BaseWsClient(CADC_AC_SERVICE_ID, Subject(),
                                 self.agent, insecure=insecure,
                                 retry=True, host=self.host)
            login_url = login._get_url((CADC_LOGIN_CAPABILITY, None))
            realm = urlparse(login_url).hostname
            auth = subject.get_auth(realm)
            if not auth:
                raise RuntimeError(
                    'No user/password for realm {} in .netrc'.format(realm))
            data = urlencode([('username', auth[0]), ('password', auth[1])])
            headers = {
                "Content-type": "application/x-www-form-urlencoded",
                "Accept": "text/plain"
            }
            cookie_response = \
                login.post((CADC_LOGIN_CAPABILITY, None), data=data,
                           headers=headers)
            cookie_response.raise_for_status()
            for cadc_realm in CADC_REALMS:
                subject.cookies.append(
                    CookieInfo(cadc_realm, CADC_SSO_COOKIE_NAME,
                               '"{}"'.format(cookie_response.text)))

        self._gms_client = BaseWsClient(resource_id, subject, self.agent,
                                        retry=True, host=self.host,
                                        insecure=insecure)
        # double check it is a GMS server
        try:
            self._search_ep = \
                self._gms_client.caps.get_access_url(SEARCH_CAPABILITY_ID)
        except Exception as ex:
            logger.error('Not a GMS service', ex)
            raise ex

        # non-standard end points
        self._groups_ep = \
            self._search_ep[:self._search_ep.rfind('/')] + '/groups'

    def create_group(self, group):
        """ Persist the given Group """
        if group is None:
            raise ValueError("Group cannot be None.")

        writer = GroupWriter()
        xml_string = writer.write(group)
        self._gms_client.put(self._groups_ep, data=xml_string)

    def get_group(self, group_id):
        if group_id is None or group_id.strip() == '':
            raise ValueError("Group ID cannot be None or empty.")

        xml_string = self._gms_client.get('{}/{}'.format(
            self._groups_ep, group_id)).content
        reader = GroupReader()
        group = reader.read(xml_string)
        return group

    def update_group(self, group):
        """
            Persist the given existing group Group
        """
        if group is None:
            raise ValueError("Group cannot be None.")

        writer = GroupWriter()
        xml_string = writer.write(group)
        self._gms_client.post('{}/{}'.format(self._groups_ep, group.group_id),
                              data=xml_string)

    def add_user_member(self, group_id, identity):
        """Add the user to the group's user members.
        group_id - the group to add the user member
        identity - the user to add as a user member
        """

        if identity is None:
            raise ValueError("User identity cannot be None.")
        if group_id is None or group_id.strip() == '':
            raise ValueError("Group ID cannot be None or empty.")

        url = '{}/{}/{}/{}'.format(
            self._groups_ep, group_id, USER_MEMBER_PATH, identity.name)
        parameters = {'idType': identity.type,
                      'userID': identity.name,
                      'groupID': group_id}
        self._gms_client.put(url, params=parameters)

    def remove_user_member(self, group_id, identity):
        """Remove the user from the group's user members.
        group_id - the group to remove the user member
        identity - the user to remove as a user member
        """

        if identity is None:
            raise ValueError("User identity cannot be None.")
        if group_id is None or group_id.strip() == '':
            raise ValueError("Group ID cannot be None or empty.")

        url = '{}/{}/{}/{}'.format(
            self._groups_ep, group_id, USER_MEMBER_PATH, identity.name)
        parameters = {'idType': identity.type,
                      'userID': identity.name,
                      'groupID': group_id}

        self._gms_client.delete(url, params=parameters)

    def add_group_member(self, group_id, member_group_id):
        """Add the group to the group's group members.
        group_id - ID of the group to add the group member to
        member_group_id - ID of the group to add to the group members
        """
        if group_id is None or not group_id.strip():
            raise ValueError("Group ID cannot be None or empty.")
        if member_group_id is None or not member_group_id.strip():
            raise ValueError("Member Group ID cannot be None or empty.")
        url = '{}/{}/{}/{}'.format(
            self._groups_ep, group_id, GROUP_MEMBER_PATH, member_group_id)
        params = {'groupID': group_id, 'groupID2': member_group_id}
        self._gms_client.put(url, params=params)

    def remove_group_member(self, group_id, member_group_id):
        """Remove the group from the group's group members.
        group_id - ID of the group to remove the member group from
        member_group_id - ID of the group to remove from group members
        """
        if group_id is None or not group_id.strip():
            raise ValueError("Group ID cannot be None or empty.")
        if member_group_id is None or not member_group_id.strip():
            raise ValueError("Member Group ID cannot be None or empty.")
        url = '{}/{}/{}/{}'.format(
            self._groups_ep, member_group_id, GROUP_MEMBER_PATH, group_id)
        params = {'groupID': group_id,
                  'groupID2': member_group_id}
        self._gms_client.delete(url, params=params)

    def get_membership(self, role=Role('member'), group_id=None):
        """Search for user group membership, of a certain role.
        role -- Role
        group_id -- if specified search only for this group
        """

        params = {}
        if group_id:
            params['groupID'] = group_id
        if role:
            params['role'] = role.get_name()
        xml_string = self._gms_client.get(
            self._search_ep, params=params).content
        reader = GroupsReader()
        groups = reader.read(xml_string, deep_copy=False)
        return groups

    def is_member(self, group_ids):
        """ Return True if user_id is a member (type of role) of at
        least one group in group_ids. False otherwise.
        group_ids -- list of group ID strings
        """

        # Allow the caller to supply a single string
        if not group_ids:
            raise ValueError('group_ids is required')
        if isinstance(group_ids, str):
            group_ids = [group_ids]

        # get_membership returns a single-element set if member of a
        # particular group. Stop as soon as we find one.
        groups = self._gms_client.get(self._search_ep,
                                      params={'group': group_ids}).text
        return len(groups.strip()) > 0

    def remove_group(self, group_id):
        """
        Remove a group from the CADC GMS
        :param group_id: The ID of the group to remove
        """
        if not group_id:
            raise ValueError('group_id is required')
        self._gms_client.delete('{}/{}'.format(self._groups_ep, group_id))

# TODO - not needed for now
#  class UsersClient(BaseClient):
#     """Class for interacting with the access control web service"""
#
#     def __init__(self, *args, **kwargs):
#         """GMS client constructor. The dn will be extracted from the
#         x509 cert and available as a default for user_id in other
#         method calls.
#         certfile -- Path to CADC proxy certificate
#         """
#
#         # This client does not support name/password authentication
#         super(UsersClient, self).__init__(usenetrc=False, *args, **kwargs)
#
#         # Specific base_url for AC webservice
#         host = os.getenv('AC_WEBSERVICE_HOST', self.host)
#         path = os.getenv('AC_WEBSERVICE_PATH', '/ac')
#         self.base_url = '%s://%s%s' % ('https', host, path)
#         self.logger.info('Base URL ' + self.base_url)
#
#         # This client will need the user DN
#         self.current_user_dn = self.get_current_user_dn()
#
#         # Specialized exceptions handled by this client
#         self._HTTP_STATUS_CODE_EXCEPTIONS[404] = {
#             "User": exceptions.UserNotFoundException(),
#             "Group": exceptions.GroupNotFoundException()
#         }
#         self._HTTP_STATUS_CODE_EXCEPTIONS[409] = \
#             exceptions.GroupExistsException()
#
#     def get_user(self, identity):
#
#         if identity is None:
#             raise ValueError("User Identity cannot be None.")
#         if not isinstance(identity, Identity):
#             raise ValueError("identity must be of type Identity.")
#
#         url = "{0}/users/{1}?idType={2}".format(self.base_url,
#         identity.name, identity.type)
#         xml_string = self._download_xml(url)
#         reader = UserReader()
#         user = reader.read(xml_string)
#         self.logger.info('Retrieved user {0}'.format(user.internal_id))
#         return user


def print_group(group):
    print('Group:')
    print('             ID: {}'.format(group.group_id))
    print('    Description: {}'.format(group.description))
    print('          Owner: {}'.format(group.owner.identities['X500'].name))
    print('  Last Modified: {}'.format(group.last_modified))
    print('    User Admins: {}'.format((',\n'+' '*17).join(
        ['{} ({})'.format(
            a.identities['HTTP'].name,
            a.identities['X500'].name) for a in group.user_admins])))
    print('   Group Admins: {}'.format(
        (',\n'+' '*17).join([gr.group_id for gr in group.group_admins])))
    print_group_members(group)


def print_group_members(group):
    print('   User Members: {}'.format((',\n'+' '*17).join(
        ['{} ({})'.format(
            m.identities['HTTP'].name,
            m.identities['X500'].name) for m in group.user_members])))
    print('  Group Members: {}'.format(
        (',\n'+' '*17).join([gr.group_id for gr in group.group_members])))


def main_app():
    parser = util.get_base_parser(version=version.version, auth_required=True,
                                  default_resource_id=CADC_AC_SERVICE_ID)

    parser.description = (
        'Client for accessing the User Group Management System (GMS) at the '
        'Canadian Astronomy Data Centre')

    subparsers = parser.add_subparsers(
        dest='cmd',
        help='supported commands. Use the -h|--help argument of a command '
             'for more details')

    list_parser = subparsers.add_parser(
        'list',
        description='Retrieve name of all the groups the user has access to',
        help='Retrieve names of all the groups the user has access to')
    list_parser.add_argument(
        '-r', '--role', choices=['admin', 'member', 'owner'],
        help='Filter by role in the group',
        required=False)
    list_parser.epilog = (
        'Example:\n'
        '        cadc-groups list --cert ~/.ssl/cadcproxy.pem\n')

    get_parser = subparsers.add_parser(
        'get',
        description='Retrieve information regarding CADC groups',
        help='Retrieve information regarding CADC groups')
    get_parser.add_argument(
        'groupid', help='the id of the CADC group', nargs='+')
    get_parser.epilog = (
        'Example:\n'
        '        cadc-groups get --cert $HOME/.ssl/cadcproxy.pem FOO\n')

    create_parser = subparsers.add_parser(
        'create',
        description='Creates a new CADC Group. Caller is owner and member of '
                    'the new group',
        help='Creates a new CADC Group. Caller is owner and member of the new '
             'group')
    create_parser.add_argument(
        '--description',
        default=None,
        help='group description',
        required=False)
    create_parser.add_argument(
        'groupid',
        help='ID of the CADC group. ID cannot contain space ( ), slash (/), '
             'escape (\\), or percent characters')
    create_parser.epilog = (
        'Examples:\n'
        '        cadc-groups create --cert ~/.ssl/cadcproxy.pem -d "My first '
        'group" -m "cn=abc" foo\n')

    members_parser = subparsers.add_parser(
        'members',
        description='Display, sets or deletes group membership information',
        help='Display, sets or deletes group membership information')
    cmd_group = members_parser.add_mutually_exclusive_group()
    cmd_group.add_argument('-clear', '--clear', action='store_true',
                           help='Clear group and user membership')
    cmd_group.add_argument('--remove-group', action='append',
                           help='Remove member group with provided ID')
    cmd_group.add_argument('--add-group', action='append',
                           help='Add member group with provided ID')
    cmd_group.add_argument('--remove-user', action='append',
                           help='Remove member user with CADC username or '
                                'distinguished name (DN)')
    cmd_group.add_argument('--add-user', action='append',
                           help='Add member user with with given CADC username'
                                ' or distinguished name (DN)')
    members_parser.add_argument('groupid', help='ID of the group')
    members_parser.epilog = (
        'Examples:\n'
        '        cadc-groups members --erase GEMINI foo\n')

    remove_parser = subparsers.add_parser(
        'remove',
        description='Remove a CADC group.',
        help='Remove a CADC group')
    remove_parser.add_argument('groupid', help='ID of the group',
                               nargs='+')

    def handle_error(exception, exit_after=True):
        """
        Prints error message and exit (by default)
        :param exception: error
        :param exit_after: True if log error message and exit,
        False if log error message and return
        :return:
        """

        if isinstance(exception, exceptions.UnauthorizedException):
            # TODO - magic authentication
            # if subject.anon:
            #     handle_error('Operation cannot be performed anonymously. '
            #      'Use one of the available methods to authenticate')
            # else:
            print('ERROR: Unexpected authentication problem')
        elif isinstance(exception, exceptions.NotFoundException):
            print('ERROR: Not found: {}'.format(str(exception)))
        elif isinstance(exception, exceptions.ForbiddenException):
            print('ERROR: Unauthorized to perform operation')
        elif isinstance(exception, exceptions.UnexpectedException):
            print('ERROR: Unexpected server error: {}'.format(str(exception)))
        else:
            print('ERROR: {}'.format(exception))

        # if logger.isEnabledFor(logging.DEBUG):
        #     traceback.print_stack()

        if exit_after:
            sys.exit(-1)  # TODO use different error codes?

    args = parser.parse_args()
    if len(sys.argv) < 2:
        parser.print_usage(file=sys.stderr)
        sys.stderr.write("{}: error: too few arguments\n".format(APP_NAME))
        sys.exit(-1)
    if args.verbose:
        logging.basicConfig(level=logging.INFO, stream=sys.stdout)
        logger.setLevel(logging.INFO)
    elif args.debug:
        logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
        logger.setLevel(logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING, stream=sys.stdout)

    subject = Subject.from_cmd_line_args(args)

    client = GroupsClient(subject, args.resource_id, host=args.host,
                          insecure=args.insecure)
    try:
        if args.cmd == 'list':
            logger.info('list')
            if args.role:
                groups = client.get_membership(Role(args.role))
            else:
                groups = client.get_membership()
            for gr in groups:
                print(gr.group_id)
        elif args.cmd == 'get':
            logger.info('get')
            for gi in args.groupid:
                print_group(client.get_group(group_id=gi))
        elif args.cmd == 'create':
            group = Group(args.groupid)
            group.description = args.description
            client.create_group(group=group)
        elif args.cmd == 'members':
            logger.info('members')
            group = client.get_group(args.groupid)
            if args.clear:
                group.group_members.clear()
                group.user_members.clear()
                client.update_group(group)
            elif args.add_user:
                for au in args.add_user:
                    if au.strip().lower().startswith('cn='):
                        id_type = 'X500'
                    else:
                        id_type = 'HTTP'
                    client.add_user_member(
                        group_id=group.group_id,
                        identity=Identity(name=au, identity_type=id_type))
            elif args.add_group:
                for ag in args.add_group:
                    client.add_group_member(group_id=group.group_id,
                                            member_group_id=ag)
            elif args.remove_user:
                for ru in args.remove_user:
                    if ru.strip().lower().startswith('cn='):
                        id_type = 'X500'
                    else:
                        id_type = 'HTTP'
                    client.remove_user_member(
                        group_id=group.group_id,
                        identity=Identity(name=ru, identity_type=id_type))
            elif args.remove_group:
                for rg in args.remove_group:
                    client.remove_group_member(group_id=group.group_id,
                                               member_group_id=rg)
            else:
                print_group_members(group)
        elif args.cmd == 'remove':
            logger.info('remove')
            for gi in args.groupid:
                client.remove_group(group_id=gi)
        else:
            raise RuntimeError('Unknown command option: ' + args.cmd)
    except Exception as ex:
        if logger.isEnabledFor(logging.DEBUG):
            traceback.print_stack()
        handle_error(ex, exit_after=True)
    else:
        logger.info("DONE")



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
# (c) 2021.                            (c) 2021.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
# ***********************************************************************

from datetime import datetime

from cadcutils.net import Group, GroupProperty, User, Identity
from cadcutils.net.group_xml.group_reader import GroupReader
from cadcutils.net.group_xml.group_writer import GroupWriter


def test_minimal_group():
    expected = Group('groupID')
    writer = GroupWriter()
    xml_string = writer.write(expected, False)

    assert xml_string

    reader = GroupReader()
    actual = reader.read(xml_string)
    assert expected.group_id
    assert actual.group_id
    assert(actual.group_id == expected.group_id)

    assert expected.owner == actual.owner
    assert expected.description == actual.description
    assert expected.last_modified == actual.last_modified

    assert(actual.group_members == expected.group_members)
    assert(actual.user_members == expected.user_members)
    assert(actual.group_admins == expected.group_admins)
    assert(actual.user_admins == expected.user_admins)


def test_maximal_group():
    owner = User('ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-00000000000a')
    owner.identities['X500'] = Identity('cn=foo,c=ca', 'X500')
    owner.identities['OpenID'] = Identity('foo@bar.com', 'OpenID')
    owner.identities['HTTP'] = Identity('foo', 'HTTP')
    owner.identities['CADC'] = Identity('00000000-0000-0000-0000-000000000001',
                                        'CADC')

    expected = Group('groupID')
    expected.owner = owner
    expected.description = 'description'
    expected.last_modified = datetime(2014, 1, 20, 19, 45, 37, 0)
    expected.properties.add(GroupProperty('key1', 'value1', True))
    expected.properties.add(GroupProperty('key2', 'special char é', False))

    user1 = User('ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-00000000000b')
    user2 = User('ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-00000000000c')
    group_member1 = Group('groupMember1')
    group_member1.owner = user1
    group_member2 = Group('groupMember2')
    group_member2.owner = user2
    expected.group_members.add(group_member1)
    expected.group_members.add(group_member2)

    user_member1 = User(
        'ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-00000000000d')
    user_member2 = User(
        'ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-00000000000e')
    expected.user_members.add(user_member1)
    expected.user_members.add(user_member2)

    owner1 = User(
        'ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-00000000000f')
    owner2 = User(
        'ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-0000000000aa')
    group_admin1 = Group('adminMember1')
    group_admin1.owner = owner1
    group_admin2 = Group('adminMember2')
    group_admin2.owner = owner2
    expected.group_admins.add(group_admin1)
    expected.group_admins.add(group_admin2)

    user_admin1 = User(
        'ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-0000000000ab')
    user_admin2 = User(
        'ivo://cadc.nrc.ca/user?00000000-0000-0000-0000-0000000000ac')
    expected.user_admins.add(user_admin1)
    expected.user_admins.add(user_admin2)

    writer = GroupWriter()
    xml_string = writer.write(expected, True)

    assert(xml_string)
    assert(len(xml_string) > 0)

    reader = GroupReader()
    actual = reader.read(xml_string)

    assert expected.group_id
    assert actual.group_id
    assert(actual.group_id == expected.group_id)

    assert(actual.owner.internal_id == expected.owner.internal_id)
    assert(actual.owner.identities == expected.owner.identities)
    assert(actual.description == expected.description)
    assert(actual.last_modified == expected.last_modified)

    assert(actual.properties == expected.properties)
    assert(actual.group_members == expected.group_members)
    assert(actual.user_members == expected.user_members)
    assert(actual.group_admins == expected.group_admins)
    assert(actual.user_admins == expected.user_admins)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
# (c) 2021.                            (c) 2021.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
# ***********************************************************************

import sys
import pytest
import os
from mock import Mock, patch, call

from cadcutils.net import Subject, GroupsClient, Group, Identity, Role
from cadcutils.net.groups_client import main_app
from cadcutils.net.group_xml import GroupWriter, GroupsWriter


THIS_DIR = os.path.dirname(os.path.realpath(__file__))
TESTDATA_DIR = os.path.join(THIS_DIR, 'data')


class MyExitError(Exception):
    def __init__(self):
        self.message = "MyExitError"


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_group_create(basews_mock):
    put_mock = Mock()
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.put = put_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    group_id = 'grID'
    gw = GroupWriter()
    group = Group(group_id=group_id)
    gr_xml = gw.write(group)
    test_client = GroupsClient(Subject())
    test_client.create_group(group)
    put_mock.assert_called_with(gms_service_url, data=gr_xml)

    with pytest.raises(ValueError):
        test_client.create_group(None)


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_group_get(basews_mock):
    group_id = 'grID'
    gw = GroupWriter()
    group = Group(group_id=group_id)
    gr_xml = gw.write(group)
    get_response = Mock(content=gr_xml)
    get_mock = Mock(return_value=get_response)
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.get = get_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    test_client = GroupsClient(Subject())
    actual_gr = test_client.get_group(group.group_id)
    get_mock.assert_called_with('{}/{}'.format(gms_service_url, group_id))
    assert group == actual_gr

    with pytest.raises(ValueError):
        test_client.get_group('')


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_group_update(basews_mock):
    post_mock = Mock()
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.post = post_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    group_id = 'grID'
    gw = GroupWriter()
    group = Group(group_id=group_id)
    gr_xml = gw.write(group)
    test_client = GroupsClient(Subject())
    test_client.update_group(group)
    post_mock.assert_called_with(
        '{}/{}'.format(gms_service_url, group.group_id), data=gr_xml)

    with pytest.raises(ValueError):
        test_client.update_group(group=None)


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_group_remove(basews_mock):
    delete_mock = Mock()
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.delete = delete_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    group_id = 'grID'
    test_client = GroupsClient(Subject())
    test_client.remove_group(group_id)
    delete_mock.assert_called_with(
        '{}/{}'.format(gms_service_url, group_id))

    with pytest.raises(ValueError):
        test_client.remove_group(None)


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_add_member(basews_mock):
    put_mock = Mock()
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.put = put_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    group_id = 'grID'
    user_id = 'userID'
    test_client = GroupsClient(Subject())
    test_client.add_user_member(
        identity=Identity(name=user_id, identity_type='CADC'),
        group_id=group_id)
    params = {'idType': 'CADC', 'userID': user_id, 'groupID': group_id}
    put_mock.assert_called_with(
        '{}/{}/userMembers/{}'.format(gms_service_url, group_id, user_id),
        params=params)

    with pytest.raises(ValueError):
        test_client.add_user_member(identity=Identity(user_id, 'CADC'),
                                    group_id='   ')

    with pytest.raises(ValueError):
        test_client.add_user_member(identity=None, group_id='abc')


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_remove_member(basews_mock):
    delete_mock = Mock()
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.delete = delete_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    group_id = 'grID'
    user_id = 'userID'
    test_client = GroupsClient(Subject())
    test_client.remove_user_member(
        identity=Identity(name=user_id, identity_type='CADC'),
        group_id=group_id)
    params = {'idType': 'CADC', 'userID': user_id, 'groupID': group_id}
    delete_mock.assert_called_with(
        '{}/{}/userMembers/{}'.format(gms_service_url, group_id, user_id),
        params=params)
    with pytest.raises(ValueError):
        test_client.remove_user_member(identity=None, group_id='abc')
    with pytest.raises(ValueError):
        test_client.remove_user_member(identity='abc', group_id='')


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_add_group_member(basews_mock):
    put_mock = Mock()
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.put = put_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    group_id = 'grID'
    member_group_id = 'memberGrID'
    test_client = GroupsClient(Subject())
    params = {'groupID': group_id, 'groupID2': member_group_id}
    test_client.add_group_member(group_id=group_id,
                                 member_group_id=member_group_id)
    put_mock.assert_called_with(
        '{}/{}/groupMembers/{}'.format(gms_service_url,
                                       group_id, member_group_id),
        params=params)

    with pytest.raises(ValueError):
        test_client.add_group_member(None, member_group_id='abc')
    with pytest.raises(ValueError):
        test_client.add_group_member(group_id='abc', member_group_id=None)


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_remove_group_member(basews_mock):
    delete_mock = Mock()
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.delete = delete_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    group_id = 'grID'
    member_group_id = 'memberGrID'
    test_client = GroupsClient(Subject())
    test_client.remove_group_member(group_id=group_id,
                                    member_group_id=member_group_id)
    params = {'groupID': group_id, 'groupID2': member_group_id}
    delete_mock.assert_called_with(
        '{}/{}/groupMembers/{}'.format(gms_service_url,
                                       member_group_id, group_id),
        params=params)

    with pytest.raises(ValueError):
        test_client.remove_group_member(group_id=None, member_group_id='abc')
    with pytest.raises(ValueError):
        test_client.remove_group_member(group_id='abc', member_group_id=None)


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_get_membership(basews_mock):
    group_id1 = 'grID1'
    group1 = Group(group_id=group_id1)
    group_id2 = 'grID2'
    group2 = Group(group_id=group_id2)
    gw = GroupsWriter()
    grs_xml = gw.write([group1, group2])
    get_response = Mock(content=grs_xml)
    get_mock = Mock(return_value=get_response)
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.get = get_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    test_client = GroupsClient(Subject())
    actual_grs = test_client.get_membership()
    get_mock.assert_called_with(gms_service_url,
                                params={'role': 'member'})
    assert 2 == len(actual_grs)
    assert group1 in actual_grs
    assert group2 in actual_grs

    get_mock.reset_mock()
    test_client.get_membership(role=Role('admin'), group_id=group_id2)
    get_mock.assert_called_with(
        gms_service_url,
        params={'role': 'admin', 'groupID': group_id2})
    assert 2 == len(actual_grs)

    get_mock.reset_mock()
    test_client.get_membership(role=Role('owner'))
    get_mock.assert_called_with(gms_service_url, params={'role': 'owner'})


@patch('cadcutils.net.groups_client.BaseWsClient')
def test_is_member(basews_mock):
    group_id = 'grID'
    get_response = Mock(text=group_id + '\n')
    get_mock = Mock(return_value=get_response)
    gms_service_url = 'https://service.org/groups'
    basews_mock.return_value.get = get_mock
    basews_mock.return_value.caps.get_access_url.return_value = gms_service_url
    test_client = GroupsClient(Subject())
    assert test_client.is_member(group_ids=group_id)
    get_mock.assert_called_with(gms_service_url,
                                params={'group': [group_id]})
    # repeat for no groups
    get_mock.reset_mock()
    get_response = Mock(text='')  # no groups response
    get_mock = Mock(return_value=get_response)
    basews_mock.return_value.get = get_mock
    assert not test_client.is_member(group_ids=['gr1', 'gr2'])
    get_mock.assert_called_with(gms_service_url,
                                params={'group': ['gr1', 'gr2']})

    with pytest.raises(ValueError):
        test_client.is_member(None)


@patch('cadcutils.net.groups_client.Subject.from_cmd_line_args',
       Mock(return_value=Subject()))
@patch('cadcutils.net.groups_client.BaseWsClient')
@patch('cadcutils.net.GroupsClient.get_group')
def test_mainapp_get(get_group_mock, basews_mock):
    basews_mock.return_value.caps.get_access_url.return_value = \
        'https://serv.com/gms'
    sys.argv = 'cadc-groups get --cert cert.pem grID'.split()
    main_app()
    get_group_mock.assert_called_with(group_id='grID')

    # multiple groups
    get_group_mock.reset_mock()
    sys.argv = 'cadc-groups get --cert cert.pem grID1 grID2'.split()
    main_app()
    get_group_mock.has_calls([call(group_id='grID1'), call(group_id='grID2')])


@patch('cadcutils.net.groups_client.Subject.from_cmd_line_args',
       Mock(return_value=Subject()))
@patch('cadcutils.net.groups_client.BaseWsClient')
@patch('cadcutils.net.GroupsClient.create_group')
def test_mainapp_create(create_group_mock, basews_mock):
    basews_mock.return_value.caps.get_access_url.return_value = \
        'https://serv.com/gms'
    sys.argv = 'cadc-groups create --cert cert.pem --description Test_gr ' \
               'grID'.split()
    main_app()
    created_group = Group(group_id='grID')
    created_group.description = 'Test_gr'
    create_group_mock.assert_called_with(group=created_group)


@patch('cadcutils.net.groups_client.Subject.from_cmd_line_args',
       Mock(return_value=Subject()))
@patch('cadcutils.net.groups_client.BaseWsClient')
@patch('cadcutils.net.GroupsClient.remove_group')
def test_mainapp_remove(remove_group_mock, basews_mock):
    basews_mock.return_value.caps.get_access_url.return_value = \
        'https://serv.com/gms'
    sys.argv = 'cadc-groups remove --cert cert.pem grID'.split()
    main_app()
    remove_group_mock.assert_called_with(group_id='grID')


@patch('cadcutils.net.groups_client.Subject.from_cmd_line_args',
       Mock(return_value=Subject()))
@patch('cadcutils.net.groups_client.BaseWsClient')
@patch('cadcutils.net.GroupsClient.get_membership')
def test_mainapp_list(get_membership_mock, basews_mock):
    basews_mock.return_value.caps.get_access_url.return_value = \
        'https://serv.com/gms'
    sys.argv = 'cadc-groups list --cert cert.pem'.split()
    main_app()
    get_membership_mock.assert_called_with()

    get_membership_mock.reset_mock()
    sys.argv = 'cadc-groups list --cert cert.pem --role admin'.split()
    main_app()
    get_membership_mock.assert_called_with(Role('admin'))


@patch('cadcutils.net.groups_client.Subject.from_cmd_line_args',
       Mock(return_value=Subject()))
@patch('cadcutils.net.groups_client.BaseWsClient')
@patch('cadcutils.net.GroupsClient.get_group')
@patch('cadcutils.net.GroupsClient.add_group_member')
@patch('cadcutils.net.GroupsClient.add_user_member')
@patch('cadcutils.net.GroupsClient.remove_group_member')
@patch('cadcutils.net.GroupsClient.remove_user_member')
def test_mainapp_members(
        remove_user_member_mock,
        remove_group_member_mock,
        add_user_member_mock,
        add_group_member_mock,
        get_group_mock,
        basews_mock):
    basews_mock.return_value.caps.get_access_url.return_value = \
        'https://serv.com/gms'
    target_grid = 'GrID'
    get_group_mock.return_value = Group(group_id=target_grid)

    # add group member
    sys.argv = \
        'cadc-groups members --cert cert.pem --add-group ABC GrID'.split()
    main_app()
    add_group_member_mock.assert_called_with(group_id=target_grid,
                                             member_group_id='ABC')

    # add user member with CADC username
    sys.argv = \
        'cadc-groups members --cert cert.pem --add-user abc GrID'.split()
    main_app()
    add_user_member_mock.assert_called_with(
        group_id=target_grid,
        identity=Identity(name='abc', identity_type='HTTP'))

    # add user member with DN
    add_user_member_mock.reset_mock()
    sys.argv = 'cadc-groups members --cert cert.pem ' \
               '--add-user CN=abc,OU=org GrID'.split()
    main_app()
    add_user_member_mock.assert_called_with(
        group_id=target_grid,
        identity=Identity(name='CN=abc,OU=org', identity_type='X500'))

    # add group member
    sys.argv = \
        'cadc-groups members --cert cert.pem --remove-group ABC GrID'.split()
    main_app()
    remove_group_member_mock.assert_called_with(group_id=target_grid,
                                                member_group_id='ABC')

    # remove user member with CADC username
    sys.argv = \
        'cadc-groups members --cert cert.pem --remove-user abc GrID'.split()
    main_app()
    remove_user_member_mock.assert_called_with(
        group_id=target_grid,
        identity=Identity(name='abc', identity_type='HTTP'))

    # add user member with DN
    add_user_member_mock.reset_mock()
    sys.argv = 'cadc-groups members --cert cert.pem ' \
               '--remove-user CN=abc,OU=org GrID'.split()
    main_app()
    remove_user_member_mock.assert_called_with(
        group_id=target_grid,
        identity=Identity(name='CN=abc,OU=org', identity_type='X500'))



# -*- coding: utf-8 -*-
# ************************************************************************
# *******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# **************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
# *
# *  (c) 2021.                            (c) 2021.
# *  Government of Canada                 Gouvernement du Canada
# *  National Research Council            Conseil national de recherches
# *  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
# *  All rights reserved                  Tous droits réservés
# *
# *  NRC disclaims any warranties,        Le CNRC dénie toute garantie
# *  expressed, implied, or               énoncée, implicite ou légale,
# *  statutory, of any kind with          de quelque nature que ce
# *  respect to the software,             soit, concernant le logiciel,
# *  including without limitation         y compris sans restriction
# *  any warranty of merchantability      toute garantie de valeur
# *  or fitness for a particular          marchande ou de pertinence
# *  purpose. NRC shall not be            pour un usage particulier.
# *  liable in any event for any          Le CNRC ne pourra en aucun cas
# *  damages, whether direct or           être tenu responsable de tout
# *  indirect, special or general,        dommage, direct ou indirect,
# *  consequential or incidental,         particulier ou général,
# *  arising from the use of the          accessoire ou fortuit, résultant
# *  software.  Neither the name          de l'utilisation du logiciel. Ni
# *  of the National Research             le nom du Conseil National de
# *  Council of Canada nor the            Recherches du Canada ni les noms
# *  names of its contributors may        de ses  participants ne peuvent
# *  be used to endorse or promote        être utilisés pour approuver ou
# *  products derived from this           promouvoir les produits dérivés
# *  software without specific prior      de ce logiciel sans autorisation
# *  written permission.                  préalable et particulière
# *                                       par écrit.
# *
# *  This file is part of the             Ce fichier fait partie du projet
# *  OpenCADC project.                    OpenCADC.
# *
# *  OpenCADC is free software:           OpenCADC est un logiciel libre ;
# *  you can redistribute it and/or       vous pouvez le redistribuer ou le
# *  modify it under the terms of         modifier suivant les termes de
# *  the GNU Affero General Public        la “GNU Affero General Public
# *  License as published by the          License” telle que publiée
# *  Free Software Foundation,            par la Free Software Foundation
# *  either version 3 of the              : soit la version 3 de cette
# *  License, or (at your option)         licence, soit (à votre gré)
# *  any later version.                   toute version ultérieure.
# *
# *  OpenCADC is distributed in the       OpenCADC est distribué
# *  hope that it will be useful,         dans l’espoir qu’il vous
# *  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
# *  without even the implied             GARANTIE : sans même la garantie
# *  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
# *  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
# *  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
# *  General Public License for           Générale Publique GNU Affero
# *  more details.                        pour plus de détails.
# *
# *  You should have received             Vous devriez avoir reçu une
# *  a copy of the GNU Affero             copie de la Licence Générale
# *  General Public License along         Publique GNU Affero avec
# *  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
# *  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
# *                                       <http://www.gnu.org/licenses/>.
# *
# ************************************************************************

import pytest
import os
import sys

from cadcutils.net import Subject, GroupsClient, Group
from cadcutils.net.groups_client import main_app

REG_HOST = 'ws.cadc-ccda.hia-iha.nrc-cnrc.gc.ca'

THIS_DIR = os.path.dirname(os.path.realpath(__file__))
TESTDATA_DIR = os.path.join(THIS_DIR, 'data')

# this must be CadcAuthtest1 cert
CADC_TESTCERT = os.environ.get('CADC_TESTCERT', '')

# ID of the test group
TEST_GROUP_ID = 'cadc-groups-int-test'
TEST_MEMBER_USER_ID = 'cadcauthtest2'
TEST_MEMBER_GROUP_ID = TEST_GROUP_ID + '-member'


@pytest.mark.intTests
@pytest.mark.skipif(not os.path.isfile(CADC_TESTCERT),
                    reason='CADC test cert required (CADC_TESTCERT environ')
def test_group_client():
    subject = Subject(certificate=CADC_TESTCERT)

    # This can only be executed once
    # sys.argv = 'cadc-groups create --cert {} {}'.format(
    #     CADC_TESTCERT, TEST_GROUP_ID).split()
    # main_app()
    # sys.argv = 'cadc-groups create --cert {} {}'.format(
    #      CADC_TESTCERT, TEST_MEMBER_GROUP_ID).split()
    # main_app()

    client = GroupsClient(subject=subject, host=REG_HOST)
    test_group = client.get_group(TEST_GROUP_ID)

    if test_group.group_members or test_group.user_members:
        test_group.user_members.clear()
        test_group.group_members.clear()
    if test_group.group_admins or test_group.user_admins:
        test_group.user_admins.clear()
        test_group.group_admins.clear()
    test_group.description = 'TEST GROUP'

    client.update_group(test_group)

    test_group = client.get_group(TEST_GROUP_ID)

    assert not test_group.group_members
    assert not test_group.user_members
    assert not test_group.group_admins
    assert not test_group.user_admins
    assert test_group.description == 'TEST GROUP'

    # test description
    test_group.description = 'NEW DESCRIPTION'
    client.update_group(test_group)
    test_group = client.get_group(TEST_GROUP_ID)
    assert test_group.description == 'NEW DESCRIPTION'

    # test user and group members
    sys.argv = 'cadc-groups members --cert {} --add-user {} {}'.format(
        CADC_TESTCERT, TEST_MEMBER_USER_ID, TEST_GROUP_ID).split()
    main_app()

    sys.argv = 'cadc-groups members --cert {} --add-group {} {}'.format(
        CADC_TESTCERT, TEST_MEMBER_GROUP_ID, TEST_GROUP_ID).split()
    main_app()

    test_group = client.get_group(TEST_GROUP_ID)

    assert len(test_group.group_members) == 1
    assert len(test_group.user_members) == 1
    assert not test_group.group_admins
    assert not test_group.user_admins

    sys.argv = 'cadc-groups members --cert {} --clear {}'.format(
        CADC_TESTCERT, TEST_GROUP_ID).split()
    main_app()
    test_group = client.get_group(TEST_GROUP_ID)

    assert not test_group.group_members
    assert not test_group.user_members
    assert not test_group.group_admins
    assert not test_group.user_admins

    test_group.group_admins.add(Group(group_id=TEST_MEMBER_GROUP_ID))
    # TODO currently not supported. Need to retrieve user info first
    # test_group.user_admins.add(
    #     User(internal_id=
    #          'ivo://cadc.nrc.ca/gms?00000000-0000-0000-0000-000000000005'))

    client.update_group(group=test_group)
    test_group = client.get_group(TEST_GROUP_ID)
    assert len(test_group.group_admins) == 1
    # assert len(test_group.user_admins) == 1

    test_group.user_admins.clear()
    test_group.group_admins.clear()
    client.update_group(group=test_group)
    test_group = client.get_group(TEST_GROUP_ID)

    assert not test_group.group_members
    assert not test_group.user_members
    assert not test_group.group_admins



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
# (c) 2021.                            (c) 2021.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
# ***********************************************************************

from lxml import etree

from cadcutils.util import date2ivoa
from .. import Group
from .group_property_writer import GroupPropertyWriter
from .user_writer import UserWriter


class GroupWriter(object):

    def write(self, group, deep_copy=True, declaration=True):

        if not isinstance(group, Group):
            raise AttributeError('group is not a Group instance')

        return etree.tostring(self.get_group_element(group, deep_copy),
                              xml_declaration=declaration,
                              encoding='UTF-8',
                              pretty_print=True)

    def get_group_element(self, group, deep_copy):
        group_element = etree.Element('group')
        group_element.set('uri', group.uri)
        self._add_owner_element(group_element, group.owner)

        if deep_copy:
            if group.description is not None:
                self._add_element(group_element, group.description,
                                  'description')
            if group.last_modified is not None:
                self._add_datetime_element(group_element, group.last_modified,
                                           'lastModified')
            self._add_properties(group_element, group.properties)
            self._add_groups(group_element, group.group_members,
                             'groupMembers')
            self._add_users(group_element, group.user_members, 'userMembers')
            self._add_groups(group_element, group.group_admins, 'groupAdmins')
            self._add_users(group_element, group.user_admins, 'userAdmins')

        return group_element

    def _add_owner_element(self, group_element, owner):
        if owner is None:
            return

        user_writer = UserWriter()
        owner_element = etree.SubElement(group_element, 'owner')
        owner_element.append(user_writer.get_user_element(owner))

    def _add_properties(self, group_element, properties):
        if properties is None or not properties:
            return
        properties_element = etree.SubElement(group_element, 'properties')
        property_writer = GroupPropertyWriter()
        for property in properties:
            if property is not None:
                properties_element.append(
                    property_writer.get_property_element(property))

    def _add_groups(self, group_element, groups, tag):
        if groups is None or not groups:
            return
        members_element = etree.SubElement(group_element, tag)
        for group in groups:
            members_element.append(self.get_group_element(group, False))

    def _add_users(self, parent, users, tag):
        if users is None or not users:
            return
        user_writer = UserWriter()
        members_element = etree.SubElement(parent, tag)
        for user in users:
            members_element.append(user_writer.get_user_element(user))

    def _add_element(self, parent, text, tag):
        if text is None:
            return
        element = etree.SubElement(parent, tag)
        if isinstance(text, str):
            element.text = text
        else:
            element.text = str(text)

    def _add_datetime_element(self, parent, value, tag):
        if value is None:
            return
        element = etree.SubElement(parent, tag)
        element.text = date2ivoa(value)
